from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from pulser_simulation.simconfig import SimConfig

from qadence.backend import BackendConfiguration
from qadence.blocks.analog import Interaction

from .devices import Device


@dataclass
class Configuration(BackendConfiguration):
    # device type
    device_type: Device = Device.IDEALIZED

    # atomic spacing
    spacing: Optional[float] = None

    # sampling rate to be used for local simulations
    sampling_rate: float = 1.0

    # solver method to pass to the Qutip solver
    method_solv: str = "adams"

    # number of solver steps to pass to the Qutip solver
    n_steps_solv: float = 1e8

    # simulation configuration with optional noise options
    sim_config: Optional[SimConfig] = None

    # add modulation to the local execution
    with_modulation: bool = False

    # Use gate-level parameters
    use_gate_params = True

    # pulse amplitude on local channel
    amplitude_local: Optional[float] = None

    # pulse amplitude on global channel
    amplitude_global: Optional[float] = None

    # detuning value
    detuning: Optional[float] = None

    # interaction type
    interaction: Interaction = Interaction.NN

    def __post_init__(self) -> None:
        if self.sim_config is not None and not isinstance(self.sim_config, SimConfig):
            raise TypeError("Wrong 'sim_config' attribute type, pass a valid SimConfig object!")



from __future__ import annotations

import typing
from enum import Enum
from itertools import chain as _flatten
from typing import Generator, List, Type, TypeVar, Union, get_args

from sympy import Basic, Expr
from torch import Tensor

from qadence.blocks import (
    AbstractBlock,
    AddBlock,
    ChainBlock,
    CompositeBlock,
    KronBlock,
    ParametricBlock,
    PrimitiveBlock,
    PutBlock,
    ScaleBlock,
    TimeEvolutionBlock,
)
from qadence.blocks.analog import AnalogBlock, AnalogComposite, ConstantAnalogRotation, WaitBlock
from qadence.blocks.analog import chain as analog_chain
from qadence.blocks.analog import kron as analog_kron
from qadence.errors import NotPauliBlockError
from qadence.logger import get_logger
from qadence.parameters import Parameter

logger = get_logger(__name__)


TPrimitiveBlock = TypeVar("TPrimitiveBlock", bound=PrimitiveBlock)
TCompositeBlock = TypeVar("TCompositeBlock", bound=CompositeBlock)


def _construct(
    Block: Type[TCompositeBlock],
    args: tuple[Union[AbstractBlock, Generator, List[AbstractBlock]], ...],
) -> TCompositeBlock:
    if len(args) == 1 and isinstance(args[0], Generator):
        args = tuple(args[0])
    return Block([b for b in args])  # type: ignore [arg-type]


def chain(*args: Union[AbstractBlock, Generator, List[AbstractBlock]]) -> ChainBlock:
    """Chain blocks sequentially. On digital backends this can be interpreted
    loosely as a matrix mutliplication of blocks. In the analog case it chains
    blocks in time.

    Arguments:
        *args: Blocks to chain. Can also be a generator.

    Returns:
        ChainBlock

    Example:
    ```python exec="on" source="material-block" result="json"
    from qadence import X, Y, chain

    b = chain(X(0), Y(0))

    # or use a generator
    b = chain(X(i) for i in range(3))
    print(b)
    ```
    """
    # ugly hack to use `AnalogChain` if we are dealing only with analog blocks
    if len(args) and all(
        isinstance(a, AnalogBlock) or isinstance(a, AnalogComposite) for a in args
    ):
        return analog_chain(*args)  # type: ignore[return-value,arg-type]
    return _construct(ChainBlock, args)


def kron(*args: Union[AbstractBlock, Generator]) -> KronBlock:
    """Stack blocks vertically. On digital backends this can be intepreted
    loosely as a kronecker product of blocks. In the analog case it executes
    blocks parallel in time.

    Arguments:
        *args: Blocks to kron. Can also be a generator.

    Returns:
        KronBlock

    Example:
    ```python exec="on" source="material-block" result="json"
    from qadence import X, Y, kron

    b = kron(X(0), Y(1))

    # or use a generator
    b = kron(X(i) for i in range(3))
    print(b)
    ```
    """
    # ugly hack to use `AnalogKron` if we are dealing only with analog blocks
    if len(args) and all(
        isinstance(a, AnalogBlock) or isinstance(a, AnalogComposite) for a in args
    ):
        return analog_kron(*args)  # type: ignore[return-value,arg-type]
    return _construct(KronBlock, args)


def add(*args: Union[AbstractBlock, Generator]) -> AddBlock:
    """Sums blocks.

    Arguments:
        *args: Blocks to add. Can also be a generator.

    Returns:
        AddBlock

    Example:
    ```python exec="on" source="material-block" result="json"
    from qadence import X, Y, add

    b = add(X(0), Y(0))

    # or use a generator
    b = add(X(i) for i in range(3))
    print(b)
    ```
    """
    return _construct(AddBlock, args)


def tag(block: AbstractBlock, tag: str) -> AbstractBlock:
    block.tag = tag
    return block


def put(block: AbstractBlock, min_qubit: int, max_qubit: int) -> PutBlock:
    from qadence.transpile import reassign

    support = tuple(range(min(block.qubit_support), max(block.qubit_support) + 1))
    shifted_block = reassign(block, {i: i - min(support) for i in support})
    return PutBlock(shifted_block, tuple(range(min_qubit, max_qubit + 1)))


def primitive_blocks(block: AbstractBlock) -> List[PrimitiveBlock]:
    """Extract the primitive blocks from a `CompositeBlock`.
    In the case of an `AddBlock`, the `AddBlock` is considered primitive.

    Args:
        blocks: An Iterable of `AbstractBlock`s.
    Returns:
        List[PrimitiveBlock]
    """

    if isinstance(block, ScaleBlock):
        return primitive_blocks(block.block)

    elif isinstance(block, PrimitiveBlock):
        return [block]

    elif isinstance(block, CompositeBlock):
        return list(_flatten(*(primitive_blocks(b) for b in block.blocks)))

    else:
        raise NotImplementedError(f"Non-supported operation of type {type(block)}")


def get_pauli_blocks(block: AbstractBlock, raises: bool = False) -> List[PrimitiveBlock]:
    """Extract Pauli operations from an arbitrary input block

    Args:
        block (AbstractBlock): The input block to extract Pauli operations from
        raises (bool, optional): Raise an exception if the block contains something
            else than Pauli blocks.

    Returns:
        List[PrimitiveBlock]: List of Pauli operations
    """
    from qadence import operations

    paulis = []
    for b in primitive_blocks(block):
        if isinstance(b, get_args(operations.TPauliBlock)):
            paulis.append(b)
        else:
            if raises:
                raise NotPauliBlockError(f"{b.name} is not a Pauli operation")
            continue

    return paulis


def parameters(block: AbstractBlock) -> list[Parameter | Basic]:
    """Extract the Parameters of a block"""
    params = []
    exprs = uuid_to_expression(block).values()
    for expr in exprs:
        symbols = list(expr.free_symbols)
        if len(symbols) == 0:
            # assert expr.is_number or isinstance(expr, sympy.Matrix)
            params.append(expr)
        else:
            for s in symbols:
                params.append(s)
    return params


def uuid_to_block(block: AbstractBlock, d: dict[str, Expr] = None) -> dict[str, ParametricBlock]:
    from qadence import operations

    d = {} if d is None else d

    if isinstance(block, ScaleBlock):
        (uuid,) = block.parameters.uuids()
        d[uuid] = block
        uuid_to_block(block.block, d)

    # special analog cases should go away soon
    elif isinstance(block, (WaitBlock, ConstantAnalogRotation, operations.AnalogEntanglement)):
        for uuid in block.parameters.uuids():
            d[uuid] = block

    elif isinstance(block, CompositeBlock) or isinstance(block, AnalogComposite):
        for b in block.blocks:
            d = uuid_to_block(b, d)

    elif isinstance(block, ParametricBlock):
        if isinstance(block, TimeEvolutionBlock) and isinstance(block.generator, AbstractBlock):
            d = uuid_to_block(block.generator, d)
        for uuid in block.parameters.uuids():
            d[uuid] = block

    elif isinstance(block, PrimitiveBlock):
        return d

    else:
        raise NotImplementedError(f"'uuid_to_block' is not implemented for block: {type(block)}")

    return d


def uuid_to_expression(block: AbstractBlock) -> dict[str, Basic]:
    return {k: v.parameters._uuid_dict[k] for k, v in uuid_to_block(block).items()}


def expression_to_uuids(block: AbstractBlock) -> dict[Expr, list[str]]:
    """Creates a mapping between unique expressions and gate-level param_ids
    of blocks using that expression."""

    uuid_to_expr = uuid_to_expression(block)
    expr_to_uuid: dict[Expr, list[str]] = {}
    for uuid, expr in uuid_to_expr.items():
        expr_to_uuid.setdefault(expr, []).append(uuid)

    return expr_to_uuid


def uuid_to_eigen(block: AbstractBlock) -> dict[str, Tensor]:
    """Creates a mapping between a parametric block's param_id and its' eigenvalues.
    This method is needed for constructing the PSR rules for a given block.

    !!! warn
        Will ignore eigenvalues of AnalogBlocks that are not yet computed.
    """

    result = {}
    for uuid, b in uuid_to_block(block).items():
        if b.eigenvalues_generator is not None:
            if b.eigenvalues_generator.numel() > 0:
                result[uuid] = b.eigenvalues_generator

                # leave only angle parameter uuid with eigenvals for ConstantAnalogRotation block
                if isinstance(block, ConstantAnalogRotation):
                    break

    return result


def expressions(block: AbstractBlock) -> list[Basic]:
    """Extract the expressions sitting in the 'parameters' field of a ParametricBlock.
    Each element of 'parameters' is a sympy expression which can be a constant,
    a single parameter or an expression consisting of both symbols and constants."""
    return list(set(uuid_to_expression(block).values()))


def block_is_qubit_hamiltonian(block: AbstractBlock) -> bool:
    try:
        _ = get_pauli_blocks(block, raises=True)
        return True
    except NotPauliBlockError:
        return False


def _support_primitive_block(
    block: PrimitiveBlock, support: dict[int, set[str]]
) -> dict[int, set[str]]:
    pauli = block.name.value if isinstance(block.name, Enum) else block.name
    index = block.qubit_support[0]
    if index in support.keys():
        support[index].add(pauli)
    else:
        support[index] = set(pauli)

    return support


def _check_commutation(block: AbstractBlock, support: dict[int, set[str]] | None = None) -> dict:
    # avoid circular import

    if support is None:
        support = {}

    if isinstance(block, AddBlock) or isinstance(block, KronBlock):
        for subblock in block.blocks:
            support = dict(_check_commutation(subblock, support=support))

    elif isinstance(block, ScaleBlock):
        support = dict(_check_commutation(block.block, support=support))

    elif isinstance(block, PrimitiveBlock):
        support = dict(_support_primitive_block(block, support))

    else:
        raise TypeError("Original block was not a Pauli-based QubitHamiltonian!")

    return support


def block_is_commuting_hamiltonian(block: AbstractBlock) -> bool:
    """Check whether a Pauli block is composed by commuting set of operators

    Args:
        block (AbstractBlock): The Pauli block

    Returns:
        bool: flag which tells whether all the elements in the
            Pauli block are commuting or not
    """
    assert block_is_qubit_hamiltonian(block), "Only working for Pauli blocks"
    support = _check_commutation(block)
    for v in support.values():
        if len(v) > 1:
            return False
    return True


def get_block_by_uuid(block: AbstractBlock, uuid: str) -> ParametricBlock:
    return uuid_to_block(block)[uuid]


def get_blocks_by_expression(
    block: AbstractBlock, expr: Union[Parameter, Expr]
) -> list[AbstractBlock]:
    expr_to_uuids = expression_to_uuids(block)
    uuid_to_blk = uuid_to_block(block)
    return [uuid_to_blk[uuid] for uuid in expr_to_uuids[expr]]


def has_duplicate_vparams(block: AbstractBlock) -> bool:
    """Check if the given block has duplicated variational parameters

    Args:
        block (AbstractBlock): The block to check

    Returns:
        bool: A boolean indicating whether the block has
            duplicated parameters or not
    """
    params = parameters(block)
    non_number = [p for p in params if not p.is_number]
    trainables = [p for p in non_number if p.trainable]
    return len(set(trainables)) != len(trainables)


@typing.no_type_check
def unroll_block_with_scaling(
    block: AbstractBlock, block_list: list[AbstractBlock] = None
) -> list[tuple[AbstractBlock, Basic]]:
    """Extract a set of terms in the given block with corresponding scales

    This function takes an input block and extracts a list of operations
    with corresponding scaling factors.

    For example, consider the following block:
        b = 2. * Z(0) * Z(1) + 3. * (kron(X0), X(1)) + kron(Y(0), Y(1)))

    This function will return the following list of tuples:
        res = [
            ([Z(0) * Z(1)], 2.),
            (kron(X(0), X(1)), 3.),
            (kron(Y(0), Y(1)), 3.),
        ]

    Args:
        block (AbstractBlock): the given block
        block_list (list[AbstractBlock], optional): A list of blocks to which append the
            found terms. If None, an empty list is returned.

    Raises:
        TypeError: If the given block does not respect the expected format

    Returns:
        tuple[list[AbstractBlock], float]: A tuple with the list of blocks
            and a scaling factor in front
    """

    # Avoid circular imports.

    def _add_block(
        add_block: AddBlock, blist: list[tuple[AbstractBlock, float]]
    ) -> list[tuple[AbstractBlock, float]]:
        for b in add_block.blocks:
            blist = unroll_block_with_scaling(b, block_list=blist)
        return blist

    if block_list is None:
        block_list = []

    if isinstance(block, ScaleBlock):
        scaled_block = block.block
        scale: Expr = block.scale

        if not isinstance(scaled_block, AddBlock):
            block_list.append((block.block, scale))

        else:
            idx = len(block_list)
            block_list = _add_block(scaled_block, block_list)
            # param_id = block.param_id

            for i in range(idx, len(block_list)):
                (b, mul) = block_list[i]
                fact = scale * mul
                # make sure it gets picked up correctly in the parameters dictionary
                # if hasattr(b, 'param_id'):
                # b.param_id = param_id

                if not mul.is_number:
                    logger.warning(
                        """
Nested add block with multiple variational parameters. This might cause undefined behavior.
Consider rewriting your block as a single AddBlock instance.

For example, if you want to define a parametric observable with multiple variational
parameters, you should make sure that each parametric operation in the
AddBlock is defined separately. To make it more clear, you should write your
block in the following way:

> theta1 = VariationalParameter("theta1")
> theta2 = VariationalParameter("theta2")
>
> generator = theta1 * kron(X(0), X(1)) + theta1 * theta2 * kron(Z(2), Z(3))

and NOT this way:

> theta1 = VariationalParameter("theta1")
> theta2 = VariationalParameter("theta2")
>
> generator = theta1 * (kron(X(0), X(1)) + theta2 * kron(Z(2), Z(3)))
"""
                    )

                block_list[i] = (b, fact)

        return block_list

    elif (
        isinstance(block, KronBlock)
        or isinstance(block, ChainBlock)
        or isinstance(block, PrimitiveBlock)
    ):
        block_list.append((block, Parameter(1.0)))
        return block_list

    elif isinstance(block, AddBlock):
        return _add_block(block, block_list)

    else:
        raise TypeError(
            "Input block has an invalid type! It "
            "should be either a ScaleBlock or one of Add, Chain "
            f"and Kron blocks. Got {type(block)}."
        )


def assert_same_block(b1: AbstractBlock, b2: AbstractBlock) -> None:
    assert type(b1) == type(
        b2
    ), f"Block {b1} is not the same type ({type(b1)}) as Block {b2} ({type(b2)})"
    assert b1.name == b2.name, f"Block {b1} and {b2} don't have the same names!"
    assert (
        b1.qubit_support == b2.qubit_support
    ), f"Block {b1} and {b2} don't have the same qubit support!"
    if isinstance(b1, ParametricBlock) and isinstance(
        b2, ParametricBlock
    ):  # if the block is parametric, we can check some additional things
        assert len(b1.parameters.items()) == len(
            b2.parameters.items()
        ), f"Blocks {b1} and {b2} have differing numbers of parameters."
        for p1, p2 in zip(b1.parameters.expressions(), b2.parameters.expressions()):
            assert p1 == p2



from __future__ import annotations

from functools import reduce
from typing import Callable, TypeVar, overload

from qadence.blocks import AbstractBlock
from qadence.circuit import QuantumCircuit

BlockOrCirc = TypeVar("BlockOrCirc", AbstractBlock, QuantumCircuit)


@overload
def transpile(
    *fs: Callable[[AbstractBlock], AbstractBlock]
) -> Callable[[AbstractBlock], AbstractBlock]:
    ...


@overload
def transpile(
    *fs: Callable[[QuantumCircuit], QuantumCircuit]
) -> Callable[[QuantumCircuit], QuantumCircuit]:
    ...


def transpile(*fs: Callable) -> Callable:
    """`AbstractBlock` or `QuantumCircuit` transpilation. Compose functions that
    accept a circuit/block and returns a circuit/block.

    Arguments:
        *fs: composable functions that either map blocks to blocks
            (`Callable[[AbstractBlock], AbstractBlock]`)
            or circuits to circuits (`Callable[[QuantumCircuit], QuantumCircuit]`).

    Returns:
        Composed function.


    Examples:

    Flatten a block of nested chains and krons:
    ```python exec="on" source="material-block" result="json"
    from qadence import *
    from qadence.transpile import transpile, flatten, scale_primitive_blocks_only

    b = chain(2 * chain(chain(X(0), Y(0))), kron(kron(X(0), X(1))))
    print(b)
    print() # markdown-exec: hide

    # both flatten and scale_primitive_blocks_only are functions that accept and
    # return a block
    t = transpile(flatten, scale_primitive_blocks_only)(b)
    print(t)
    ```

    We also proved a decorator to easily turn a function `Callable[[AbstractBlock], AbstractBlock]`
    into a `Callable[[QuantumCircuit], QuantumCircuit]` to be used in circuit transpilation.
    ```python exec="on" source="material-block" result="json"
    from qadence import *
    from qadence.transpile import transpile, blockfn_to_circfn, flatten

    # We want to pass this circuit to `transpile` instead of a block,
    # so we need functions that map from a circuit to a circuit.
    circ = QuantumCircuit(2, chain(chain(X(0), chain(X(1)))))

    @blockfn_to_circfn
    def fn(block):
        # un-decorated function accepts a block and returns a block
        return block * block

    transp = transpile(
        # the decorated function accepts a circuit and returns a circuit
        fn,
        # already existing functions can also be decorated
        blockfn_to_circfn(flatten)
    )
    print(transp(circ))
    ```
    """
    return lambda x: reduce(lambda acc, f: f(acc), reversed(fs), x)


def blockfn_to_circfn(
    fn: Callable[[AbstractBlock], AbstractBlock]
) -> Callable[[QuantumCircuit], QuantumCircuit]:
    return lambda circ: QuantumCircuit(circ.register, fn(circ.block))



from __future__ import annotations

from itertools import chain as flatten
from typing import Callable, Dict, List

from braket.circuits.gates import CZ, CNot, CPhaseShift, H, I, Rx, Ry, Rz, S, Swap, T, X, Y, Z
from braket.circuits.instruction import Instruction
from braket.parametric import FreeParameter

from qadence.blocks import AbstractBlock, CompositeBlock, PrimitiveBlock
from qadence.errors import NotSupportedError
from qadence.operations import OpName
from qadence.parameters import evaluate

single_qubit: Dict[str, Callable] = {
    OpName.I: I.i,
    OpName.H: H.h,
    OpName.X: X.x,
    OpName.Y: Y.y,
    OpName.Z: Z.z,
    OpName.S: S.s,
    OpName.T: T.t,
}
single_qubit_parameterized: Dict[str, Callable] = {
    OpName.RX: Rx.rx,
    OpName.RY: Ry.ry,
    OpName.RZ: Rz.rz,
}
two_qubit: Dict[str, Callable] = {OpName.CNOT: CNot.cnot, OpName.SWAP: Swap.swap, OpName.CZ: CZ.cz}
two_qubit_parametrized: Dict[str, Callable] = {
    OpName.CPHASE: CPhaseShift.cphaseshift,
}

ops_map = {
    **single_qubit,
    **single_qubit_parameterized,
    **two_qubit,
    **two_qubit_parametrized,
}

supported_gates = list(ops_map.keys())


def BraketOperation(block: PrimitiveBlock) -> Instruction:
    operation = block.name

    if operation in single_qubit:
        return single_qubit[operation](target=block.qubit_support)

    elif operation in single_qubit_parameterized:
        ((uuid, expr),) = block.parameters.items()  # type: ignore [attr-defined]
        if expr.is_number:
            return single_qubit_parameterized[operation](
                target=block.qubit_support, angle=evaluate(expr)  # type: ignore
            )
        else:
            return single_qubit_parameterized[operation](
                target=block.qubit_support,
                angle=FreeParameter(uuid),  # type: ignore
            )

    elif operation in two_qubit:
        return two_qubit[operation](block.qubit_support[0], block.qubit_support[1])

    elif operation in two_qubit_parametrized:
        (expr,) = block.parameters.expressions()  # type: ignore [attr-defined]
        angle_value = evaluate(expr)
        return two_qubit_parametrized[operation](
            control=block.qubit_support[0],
            target=block.qubit_support[1],
            angle=angle_value,
        )

    else:
        raise NotSupportedError(
            "Operation type {} is not supported for Braket backend.".format(type(block))
        )


def convert_block(block: AbstractBlock) -> List[Instruction]:
    if isinstance(block, PrimitiveBlock):
        ops = [BraketOperation(block=block)]
    elif isinstance(block, CompositeBlock):
        ops = list(flatten(convert_block(b) for b in block.blocks))
    else:
        raise NotSupportedError(
            "Operation type {} is not supported for Braket backend.".format(type(block))
        )
    return ops



from __future__ import annotations

import warnings
from typing import List, Tuple, Type, Union

import numpy as np
import torch

from qadence.blocks import AbstractBlock, add
from qadence.logger import get_logger
from qadence.operations import N, X, Y, Z
from qadence.register import Register
from qadence.types import Interaction, TArray

logger = get_logger(__name__)


def interaction_zz(i: int, j: int) -> AbstractBlock:
    """Ising ZZ interaction."""
    return Z(i) @ Z(j)


def interaction_nn(i: int, j: int) -> AbstractBlock:
    """Ising NN interaction."""
    return N(i) @ N(j)


def interaction_xy(i: int, j: int) -> AbstractBlock:
    """XY interaction."""
    return X(i) @ X(j) + Y(i) @ Y(j)


def interaction_xyz(i: int, j: int) -> AbstractBlock:
    """Heisenberg XYZ interaction."""
    return X(i) @ X(j) + Y(i) @ Y(j) + Z(i) @ Z(j)


INTERACTION_DICT = {
    Interaction.ZZ: interaction_zz,
    Interaction.NN: interaction_nn,
    Interaction.XY: interaction_xy,
    Interaction.XYZ: interaction_xyz,
}


ARRAYS = (list, np.ndarray, torch.Tensor)

DETUNINGS = (N, X, Y, Z)

TDetuning = Union[Type[N], Type[X], Type[Y], Type[Z]]


def hamiltonian_factory(
    register: Register | int,
    interaction: Interaction | None = None,
    detuning: TDetuning | None = None,
    interaction_strength: TArray | str | None = None,
    detuning_strength: TArray | str | None = None,
    random_strength: bool = False,
    force_update: bool = False,
) -> AbstractBlock:
    """
    General Hamiltonian creation function. Can be used to create Hamiltonians with 2-qubit
    interactions and single-qubit detunings, both with arbitrary strength or parameterized.

    Arguments:
        register: register of qubits with a specific graph topology, or number of qubits.
            When passing a number of qubits a register with all-to-all connectivity
            is created.
        interaction: Interaction.ZZ, Interaction.NN, Interaction.XY, or Interacton.XYZ.
        detuning: single-qubit operator N, X, Y, or Z.
        interaction_strength: list of values to be used as the interaction strength for each
            pair of qubits. Should be ordered following the order of `Register(n_qubits).edges`.
            Alternatively, some string "x" can be passed, which will create a parameterized
            interactions for each pair of qubits, each labelled as `"x_ij"`.
        detuning_strength: list of values to be used as the detuning strength for each qubit.
            Alternatively, some string "x" can be passed, which will create a parameterized
            detuning for each qubit, each labelled as `"x_i"`.
        random_strength: set random interaction and detuning strengths between -1 and 1.
        force_update: force override register detuning and interaction strengths.

    Examples:
        ```python exec="on" source="material-block" result="json"
        from qadence import hamiltonian_factory, Interaction, Register, Z

        n_qubits = 3

        # Constant total magnetization observable:
        observable = hamiltonian_factory(n_qubits, detuning = Z)

        # Parameterized total magnetization observable:
        observable = hamiltonian_factory(n_qubits, detuning = Z, detuning_strength = "z")

        # Random all-to-all XY Hamiltonian generator:
        generator = hamiltonian_factory(
            n_qubits,
            interaction = Interaction.XY,
            random_strength = True,
            )

        # Parameterized NN Hamiltonian generator with a square grid interaction topology:
        register = Register.square(qubits_side = n_qubits)
        generator = hamiltonian_factory(
            register,
            interaction = Interaction.NN,
            interaction_strength = "theta"
            )
        ```
    """

    if interaction is None and detuning is None:
        raise ValueError("Please provide an interaction and/or detuning for the Hamiltonian.")

    # If number of qubits is given, creates all-to-all register
    register = Register(register) if isinstance(register, int) else register

    # Get interaction function
    try:
        int_fn = INTERACTION_DICT[interaction]  # type: ignore [index]
    except (KeyError, ValueError) as error:
        if interaction is None:
            pass
        else:
            raise KeyError(f"Interaction {interaction} not supported.")

    # Check single-qubit detuning
    if (detuning is not None) and (detuning not in DETUNINGS):
        raise TypeError(f"Detuning of type {type(detuning)} not supported.")

    # Pre-process detuning and interaction strengths and update register
    has_detuning_strength, detuning_strength = _preprocess_strengths(
        register, detuning_strength, "nodes", force_update, random_strength
    )
    has_interaction_strength, interaction_strength = _preprocess_strengths(
        register, interaction_strength, "edges", force_update, random_strength
    )

    if (not has_detuning_strength) or force_update:
        register = _update_detuning_strength(register, detuning_strength)

    if (not has_interaction_strength) or force_update:
        register = _update_interaction_strength(register, interaction_strength)

    # Create single-qubit detunings:
    single_qubit_terms: List[AbstractBlock] = []
    if detuning is not None:
        for node in register.nodes:
            block_sq = detuning(node)  # type: ignore [operator]
            strength_sq = register.nodes[node]["strength"]
            single_qubit_terms.append(strength_sq * block_sq)

    # Create two-qubit interactions:
    two_qubit_terms: List[AbstractBlock] = []
    if interaction is not None:
        for edge in register.edges:
            block_tq = int_fn(*edge)  # type: ignore [operator]
            strength_tq = register.edges[edge]["strength"]
            two_qubit_terms.append(strength_tq * block_tq)

    return add(*single_qubit_terms, *two_qubit_terms)


def _preprocess_strengths(
    register: Register,
    strength: TArray | str | None,
    nodes_or_edges: str,
    force_update: bool,
    random_strength: bool,
) -> Tuple[bool, Union[TArray | str]]:
    data = getattr(register, nodes_or_edges)

    # Useful for error messages:
    strength_target = "detuning" if nodes_or_edges == "nodes" else "interaction"

    # First we check if strength values already exist in the register
    has_strength = any(["strength" in data[i] for i in data])
    if has_strength and not force_update:
        if strength is not None:
            logger.warning(
                "Register already includes " + strength_target + " strengths. "
                "Skipping update. Use `force_update = True` to override them."
            )
    # Next we process the strength given in the input arguments
    if strength is None:
        if random_strength:
            strength = 2 * torch.rand(len(data), dtype=torch.double) - 1
        else:
            # None defaults to constant = 1.0
            strength = torch.ones(len(data), dtype=torch.double)
    elif isinstance(strength, ARRAYS):
        # If array is given, checks it has the correct length
        if len(strength) != len(data):
            message = "Array of " + strength_target + " strengths has incorrect size."
            raise ValueError(message)
    elif isinstance(strength, str):
        # Any string will be used as a prefix to variational parameters
        pass
    else:
        # If not of the accepted types ARRAYS or str, we error out
        raise TypeError(
            "Incorrect " + strength_target + f" strength type {type(strength)}. "
            "Please provide an array of strength values, or a string for "
            "parameterized " + strength_target + "s."
        )

    return has_strength, strength


def _update_detuning_strength(register: Register, detuning_strength: TArray | str) -> Register:
    for node in register.nodes:
        if isinstance(detuning_strength, str):
            register.nodes[node]["strength"] = detuning_strength + f"_{node}"
        elif isinstance(detuning_strength, ARRAYS):
            register.nodes[node]["strength"] = detuning_strength[node]
    return register


def _update_interaction_strength(
    register: Register, interaction_strength: TArray | str
) -> Register:
    for idx, edge in enumerate(register.edges):
        if isinstance(interaction_strength, str):
            register.edges[edge]["strength"] = interaction_strength + f"_{edge[0]}{edge[1]}"
        elif isinstance(interaction_strength, ARRAYS):
            register.edges[edge]["strength"] = interaction_strength[idx]
    return register


# FIXME: Previous hamiltonian / observable functions, now refactored, to be deprecated:

DEPRECATION_MESSAGE = "This function will be removed in the future. "


def single_z(qubit: int = 0, z_coefficient: float = 1.0) -> AbstractBlock:
    message = DEPRECATION_MESSAGE + "Please use `z_coefficient * Z(qubit)` directly."
    warnings.warn(message, FutureWarning)
    return Z(qubit) * z_coefficient


def total_magnetization(n_qubits: int, z_terms: np.ndarray | list | None = None) -> AbstractBlock:
    message = (
        DEPRECATION_MESSAGE
        + "Please use `hamiltonian_factory(n_qubits, detuning=Z, node_coeff=z_terms)`."
    )
    warnings.warn(message, FutureWarning)
    return hamiltonian_factory(n_qubits, detuning=Z, detuning_strength=z_terms)


def zz_hamiltonian(
    n_qubits: int,
    z_terms: np.ndarray | None = None,
    zz_terms: np.ndarray | None = None,
) -> AbstractBlock:
    message = (
        DEPRECATION_MESSAGE
        + """
Please use `hamiltonian_factory(n_qubits, Interaction.ZZ, Z, interaction_strength, z_terms)`. \
Note that the argument `zz_terms` in this function is a 2D array of size `(n_qubits, n_qubits)`, \
while `interaction_strength` is expected as a 1D array of size `0.5 * n_qubits * (n_qubits - 1)`."""
    )
    warnings.warn(message, FutureWarning)
    if zz_terms is not None:
        register = Register(n_qubits)
        interaction_strength = [zz_terms[edge[0], edge[1]] for edge in register.edges]
    else:
        interaction_strength = None

    return hamiltonian_factory(n_qubits, Interaction.ZZ, Z, interaction_strength, z_terms)


def ising_hamiltonian(
    n_qubits: int,
    x_terms: np.ndarray | None = None,
    z_terms: np.ndarray | None = None,
    zz_terms: np.ndarray | None = None,
) -> AbstractBlock:
    message = (
        DEPRECATION_MESSAGE
        + """
You can build a general transverse field ising model with the `hamiltonian_factory` function. \
Check the hamiltonian construction tutorial in the documentation for more information."""
    )
    warnings.warn(message, FutureWarning)
    zz_ham = zz_hamiltonian(n_qubits, z_terms=z_terms, zz_terms=zz_terms)
    x_ham = hamiltonian_factory(n_qubits, detuning=X, detuning_strength=x_terms)
    return zz_ham + x_ham



from __future__ import annotations

import importlib
from string import Template
from typing import TypeVar

from qadence.backend import Backend
from qadence.blocks import (
    AbstractBlock,
)
from qadence.types import BackendName, DiffMode

TAbstractBlock = TypeVar("TAbstractBlock", bound=AbstractBlock)

backends_namespace = Template("qadence.backends.$name")


def _available_backends() -> dict:
    res = {}
    for backend in BackendName.list():
        module_path = f"qadence.backends.{backend}.backend"
        try:
            module = importlib.import_module(module_path)
            BackendCls = getattr(module, "Backend")
            res[backend] = BackendCls
        except (ImportError, ModuleNotFoundError):
            pass

    return res


def _supported_gates(name: BackendName | str) -> list[TAbstractBlock]:
    from qadence import operations

    name = str(BackendName(name).name.lower())

    try:
        backend_namespace = backends_namespace.substitute(name=name)
        module = importlib.import_module(backend_namespace)
    except KeyError:
        pass
    _supported_gates = getattr(module, "supported_gates", None)
    assert (
        _supported_gates is not None
    ), f"{name} backend should define a 'supported_gates' variable"
    return [getattr(operations, gate) for gate in _supported_gates]


def _gpsr_fns() -> dict:
    # avoid circular import
    from qadence.backends.gpsr import general_psr

    return {DiffMode.GPSR: general_psr}


def _validate_diff_mode(backend: Backend, diff_mode: DiffMode) -> None:
    if not backend.supports_ad and diff_mode == DiffMode.AD:
        raise TypeError(f"Backend {backend.name} does not support diff_mode {DiffMode.AD}.")


def _set_backend_config(backend: Backend, diff_mode: DiffMode) -> None:
    """_summary_

    Args:
        backend (Backend): _description_
        diff_mode (DiffMode): _description_
    """

    _validate_diff_mode(backend, diff_mode)

    if not backend.supports_ad or diff_mode != DiffMode.AD:
        backend.config._use_gate_params = True

    # (1) When using PSR with any backend or (2)  we use the backends Pulser or Braket,
    # we have to use gate-level parameters

    else:
        assert diff_mode == DiffMode.AD
        backend.config._use_gate_params = False
        # We can use expression-level parameters for AD.
        if backend.name == BackendName.PYQTORCH:
            backend.config.use_single_qubit_composition = True

        # For pyqtorch, we enable some specific transpilation passes.


# if proprietary qadence_plus is available import the
# right function since more backends are supported
try:
    module = importlib.import_module("qadence_extensions.extensions")
    available_backends = getattr(module, "available_backends")
    supported_gates = getattr(module, "supported_gates")
    get_gpsr_fns = getattr(module, "gpsr_fns")
    set_backend_config = getattr(module, "set_backend_config")
except ModuleNotFoundError:
    available_backends = _available_backends
    supported_gates = _supported_gates
    get_gpsr_fns = _gpsr_fns
    set_backend_config = _set_backend_config



from __future__ import annotations

import logging
import os
import sys

logging_levels = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
}

LOG_STREAM_HANDLER = sys.stdout

DEFAULT_LOGGING_LEVEL = logging.INFO

# FIXME: introduce a better handling of the configuration
LOGGING_LEVEL = os.environ.get("LOGGING_LEVEL", "warning").upper()


def get_logger(name: str) -> logging.Logger:
    logger: logging.Logger = logging.getLogger(name)

    level = logging_levels.get(LOGGING_LEVEL, DEFAULT_LOGGING_LEVEL)
    logger.setLevel(level)

    formatter = logging.Formatter("%(levelname) -5s %(asctime)s: %(message)s", "%Y-%m-%d %H:%M:%S")
    # formatter = logging.Formatter(LOG_FORMAT)
    sh = logging.StreamHandler(LOG_STREAM_HANDLER)
    sh.setFormatter(formatter)
    logger.addHandler(sh)

    return logger



from __future__ import annotations

from collections import Counter
from functools import reduce

import numpy as np
import torch
from torch import Tensor

from qadence import BackendName, DiffMode
from qadence.backends import backend_factory
from qadence.blocks import AbstractBlock, PrimitiveBlock
from qadence.blocks.utils import unroll_block_with_scaling
from qadence.circuit import QuantumCircuit
from qadence.operations import H, SDagger, X, Y, Z, chain
from qadence.parameters import evaluate
from qadence.utils import Endianness


def get_qubit_indices_for_op(pauli_term: tuple, op: PrimitiveBlock | None = None) -> list[int]:
    """Get qubit indices for the given op in the Pauli term if any."""
    indices = []
    blocks = getattr(pauli_term[0], "blocks", None)
    if blocks is not None:
        for block in blocks:
            if op is None:
                indices.append(block.qubit_support[0])
            if isinstance(block, type(op)):
                indices.append(block.qubit_support[0])
    else:
        block = pauli_term[0]
        if op is None:
            indices.append(block.qubit_support[0])
        if isinstance(block, type(op)):
            indices.append(block.qubit_support[0])
    return indices


def rotate(circuit: QuantumCircuit, pauli_term: tuple) -> QuantumCircuit:
    """Rotate circuit to measurement basis and return the qubit support."""
    rotations = []

    # Mypy expects concrete types. Although there definitely should be
    # a better way to pass the operation type.
    for op, gate in [(X(0), Z), (Y(0), SDagger)]:
        qubit_indices = get_qubit_indices_for_op(pauli_term, op=op)
        for index in qubit_indices:
            rotations.append(gate(index) * H(index))
    rotated_block = chain(circuit.block, *rotations)
    return QuantumCircuit(circuit.register, rotated_block)


def get_counts(samples: list, support: list) -> list:
    """Marginalise the probablity mass function to the support."""
    counts = []
    for sample in samples:
        sample_counts = []
        for k, v in sample.items():
            sample_counts.append(Counter({"".join([k[i] for i in support]): sample[k]}))
        reduced_counts = reduce(lambda x, y: x + y, sample_counts)
        counts.append(reduced_counts)
    return counts


def empirical_average(samples: list, support: list) -> Tensor:
    """Compute the empirical average."""
    counters = get_counts(samples, support)
    expectations = []
    n_shots = np.sum(list(counters[0].values()))
    parity = -1
    for counter in counters:
        counter_exps = []
        for bitstring, count in counter.items():
            counter_exps.append(count * parity ** (np.sum([int(bit) for bit in bitstring])))
        expectations.append(np.sum(counter_exps) / n_shots)
    return torch.tensor(expectations)


def iterate_pauli_decomposition(
    circuit: QuantumCircuit,
    param_values: dict,
    pauli_decomposition: list,
    n_shots: int,
    state: Tensor | None = None,
    backend_name: BackendName = BackendName.PYQTORCH,
    endianness: Endianness = Endianness.BIG,
) -> Tensor:
    """Estimate total expectation value by averaging all Pauli terms."""

    estimated_values = []

    backend = backend_factory(backend=backend_name, diff_mode=DiffMode.GPSR)
    for pauli_term in pauli_decomposition:
        if pauli_term[0].is_identity:
            estimated_values.append(evaluate(pauli_term[1], as_torch=True))
        else:
            # Get the full qubit support for the Pauli term.
            # Note: duplicates must be kept here to allow for
            # observables chaining multiple operations on the same qubit
            # such as `b = chain(Z(0), Z(0))`
            support = get_qubit_indices_for_op(pauli_term)
            # Rotate the circuit according to the given observable term.
            rotated_circuit = rotate(circuit=circuit, pauli_term=pauli_term)
            # Use the low-level backend API to avoid embedding of parameters
            # already performed at the higher QuantumModel level.
            # Therefore, parameters passed here have already been embedded.
            conv_circ = backend.circuit(rotated_circuit)
            samples = backend.sample(
                circuit=conv_circ,
                param_values=param_values,
                n_shots=n_shots,
                state=state,
                endianness=endianness,
            )
            estim_values = empirical_average(samples=samples, support=support)
            # TODO: support for parametric observables to be tested
            estimated_values.append(estim_values * evaluate(pauli_term[1]))
    res = torch.sum(torch.stack(estimated_values), axis=0)
    # Allow for automatic differentiation.
    res.requires_grad = True
    return res


def compute_expectation(
    circuit: QuantumCircuit,
    observables: list[AbstractBlock],
    param_values: dict,
    options: dict,
    state: Tensor | None = None,
    backend_name: BackendName = BackendName.PYQTORCH,
    endianness: Endianness = Endianness.BIG,
) -> Tensor:
    """Basic tomography protocol with rotations

    Given a circuit and a list of observables, apply basic tomography protocol to estimate
    the expectation values.
    """
    if not isinstance(observables, list):
        raise TypeError(
            "Observables must be of type <class 'List[AbstractBlock]'>. Got {}.".format(
                type(observables)
            )
        )
    n_shots = options.get("n_shots")
    if n_shots is None:
        raise KeyError("Tomography protocol requires a 'n_shots' kwarg of type 'int'.")
    estimated_values = []
    for observable in observables:
        pauli_decomposition = unroll_block_with_scaling(observable)
        estimated_values.append(
            iterate_pauli_decomposition(
                circuit=circuit,
                param_values=param_values,
                pauli_decomposition=pauli_decomposition,
                n_shots=n_shots,
                state=state,
                backend_name=backend_name,
                endianness=endianness,
            )
        )
    return torch.transpose(torch.vstack(estimated_values), 1, 0)



from __future__ import annotations

from functools import singledispatch
from typing import overload

from qadence import HamEvo, QuantumCircuit
from qadence.blocks import AbstractBlock, CompositeBlock
from qadence.blocks.utils import _construct
from qadence.operations import U
from qadence.types import LTSOrder


@overload
def digitalize(circuit: QuantumCircuit, approximation: LTSOrder = LTSOrder.BASIC) -> QuantumCircuit:
    ...


@overload
def digitalize(block: AbstractBlock, approximation: LTSOrder = LTSOrder.BASIC) -> AbstractBlock:
    ...


@singledispatch
def digitalize(
    circ_or_block: AbstractBlock | QuantumCircuit, approximation: LTSOrder
) -> AbstractBlock | QuantumCircuit:
    raise NotImplementedError(f"digitalize is not implemented for {type(circ_or_block)}")


@digitalize.register  # type: ignore[attr-defined]
def _(block: AbstractBlock, approximation: LTSOrder = LTSOrder.BASIC) -> AbstractBlock:
    if isinstance(block, CompositeBlock):
        return _construct(type(block), tuple(digitalize(b, approximation) for b in block.blocks))
    elif isinstance(block, HamEvo):
        return block.digital_decomposition(approximation=approximation)
    elif isinstance(block, U):
        return block.digital_decomposition()
    else:
        return block


@digitalize.register  # type: ignore[attr-defined]
def _(circuit: QuantumCircuit, approximation: LTSOrder = LTSOrder.BASIC) -> QuantumCircuit:
    return QuantumCircuit(circuit.n_qubits, digitalize(circuit.block, approximation))



from __future__ import annotations

from collections import Counter
from typing import Sequence

import numpy as np
import torch
from torch import Tensor

from qadence.utils import Endianness, int_to_basis

# Dict of NumPy dtype -> torch dtype (when the correspondence exists)
numpy_to_torch_dtype_dict = {
    np.bool_: torch.bool,
    np.uint8: torch.uint8,
    np.int8: torch.int8,
    np.int16: torch.int16,
    np.int32: torch.int32,
    np.int64: torch.int64,
    np.float16: torch.float16,
    np.float32: torch.float32,
    np.float64: torch.float64,
    np.complex64: torch.complex64,
    np.complex128: torch.complex128,
    int: torch.int64,
    float: torch.float64,
    complex: torch.complex128,
}


def param_dict(keys: Sequence[str], values: Sequence[Tensor]) -> dict[str, Tensor]:
    return {key: val for key, val in zip(keys, values)}


def numpy_to_tensor(
    x: np.ndarray,
    device: torch.device = torch.device("cpu"),
    dtype: torch.dtype = torch.complex128,
    requires_grad: bool = False,
) -> Tensor:
    """This only copies the numpy array if device or dtype are different than the ones of x."""
    return torch.as_tensor(x, dtype=dtype, device=device).requires_grad_(requires_grad)


def promote_to_tensor(
    x: Tensor | np.ndarray | float,
    dtype: torch.dtype = torch.complex128,
    requires_grad: bool = True,
) -> Tensor:
    """Convert the given type inco a torch.Tensor"""
    if isinstance(x, float):
        return torch.tensor([[x]], dtype=dtype, requires_grad=requires_grad)
    elif isinstance(x, np.ndarray):
        return numpy_to_tensor(
            x, dtype=numpy_to_torch_dtype_dict.get(x.dtype), requires_grad=requires_grad
        )
    elif isinstance(x, Tensor):
        return x.requires_grad_(requires_grad)
    else:
        raise ValueError(f"Don't know how to promote {type(x)} to Tensor")


# FIXME: Not being used, maybe remove in v1.0.0
def count_bitstrings(sample: Tensor, endianness: Endianness = Endianness.BIG) -> Counter:
    # Convert to a tensor of integers.
    n_qubits = sample.size()[1]
    base = torch.ones(n_qubits, dtype=torch.int64) * 2
    powers_of_2 = torch.pow(base, reversed(torch.arange(n_qubits)))
    int_tensor = torch.matmul(sample, powers_of_2)
    # Count occurences of integers.
    count_int = torch.bincount(int_tensor)
    # Return a Counter for non-empty bitstring counts.
    return Counter(
        {
            int_to_basis(k=k, n_qubits=n_qubits, endianness=endianness): count.item()
            for k, count in enumerate(count_int)
            if count > 0
        }
    )


def to_list_of_dicts(param_values: dict[str, Tensor]) -> list[dict[str, float]]:
    if not param_values:
        return [param_values]

    max_batch_size = max(p.size()[0] for p in param_values.values())
    batched_values = {
        k: (v if v.size()[0] == max_batch_size else v.repeat(max_batch_size, 1))
        for k, v in param_values.items()
    }

    return [{k: v[i] for k, v in batched_values.items()} for i in range(max_batch_size)]



"""Basic operations to be implemented by backends."""
from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass
from functools import cached_property
from typing import Any, Tuple, Union

import numpy as np
import sympy
import torch
from rich.console import Console, RenderableType
from rich.padding import Padding
from rich.panel import Panel
from rich.tree import Tree
from sympy import Basic
from torch import Tensor, cdouble, tensor

from qadence.blocks import (
    AbstractBlock,
    ControlBlock,
    ParametricBlock,
    ParametricControlBlock,
    PrimitiveBlock,
    TimeEvolutionBlock,
)
from qadence.blocks.analog import (
    AnalogBlock,
    ConstantAnalogRotation,
    QubitSupport,
    WaitBlock,
)
from qadence.blocks.block_to_tensor import block_to_tensor
from qadence.blocks.utils import (
    add,  # noqa
    block_is_commuting_hamiltonian,
    block_is_qubit_hamiltonian,
    chain,
    expressions,
    kron,
)
from qadence.decompose import lie_trotter_suzuki
from qadence.logger import get_logger
from qadence.parameters import (
    Parameter,
    ParamMap,
    evaluate,
    extract_original_param_entry,
)
from qadence.types import LTSOrder, OpName, TGenerator, TNumber, TParameter
from qadence.utils import eigenvalues

logger = get_logger(__name__)


# Modules to be automatically added to the qadence namespace
__all__ = [
    "X",
    "Y",
    "Z",
    "N",
    "H",
    "I",
    "Zero",
    "RX",
    "RY",
    "RZ",
    "U",
    "CNOT",
    "CZ",
    "MCZ",
    "HamEvo",
    "CRX",
    "MCRX",
    "CRY",
    "MCRY",
    "CRZ",
    "MCRZ",
    "T",
    "TDagger",
    "S",
    "SDagger",
    "SWAP",
    "PHASE",
    "CPHASE",
    "MCPHASE",
    "wait",
    "entangle",
    "AnalogEntanglement",
    "AnalogRot",
    "AnalogRX",
    "AnalogRY",
    "AnalogRZ",
    "AnalogSWAP",
]


class X(PrimitiveBlock):
    """The X gate"""

    name = OpName.X

    def __init__(self, target: int):
        super().__init__((target,))

    @property
    def generator(self) -> AbstractBlock:
        return self

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([-1, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return tensor([-1, 1], dtype=cdouble)

    def dagger(self) -> X:
        return self


class Y(PrimitiveBlock):
    """The Y gate"""

    name = OpName.Y

    def __init__(self, target: int):
        super().__init__((target,))

    @property
    def generator(self) -> AbstractBlock:
        return self

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([-1, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return tensor([-1, 1], dtype=cdouble)

    def dagger(self) -> Y:
        return self


class Z(PrimitiveBlock):
    """The Z gate"""

    name = OpName.Z

    def __init__(self, target: int):
        super().__init__((target,))

    @property
    def generator(self) -> AbstractBlock:
        return self

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([-1, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return tensor([-1, 1], dtype=cdouble)

    def dagger(self) -> Z:
        return self


class N(PrimitiveBlock):
    """The N = (1/2)(I-Z) operator"""

    name = OpName.N

    def __init__(self, target: int):
        super().__init__((target,))

    @property
    def generator(self) -> None:
        raise ValueError("Property `generator` not available for non-unitary operator.")
        return None

    @property
    def eigenvalues_generator(self) -> None:
        raise ValueError("Property `eigenvalues_generator` not available for non-unitary operator.")
        return None

    @property
    def eigenvalues(self) -> Tensor:
        return tensor([0, 1], dtype=cdouble)

    def dagger(self) -> N:
        return self


class S(PrimitiveBlock):
    """The S / Phase gate"""

    name = OpName.S

    def __init__(self, target: int):
        self.generator = I(target) - Z(target)
        super().__init__((target,))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([0, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return tensor([1, 1j], dtype=cdouble)

    def dagger(self) -> SDagger:
        return SDagger(*self.qubit_support)


class SDagger(PrimitiveBlock):
    """The Hermitian adjoint/conjugate transpose of the S / Phase gate"""

    name = OpName.SDAGGER

    def __init__(self, target: int):
        self.generator = I(target) - Z(target)
        super().__init__((target,))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([0, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return tensor([1, -1j], dtype=cdouble)

    def dagger(self) -> S:
        return S(*self.qubit_support)


class PHASE(ParametricBlock):
    """The Parametric Phase / S gate"""

    name = OpName.PHASE

    def __init__(self, target: int, parameter: Parameter | TNumber | sympy.Expr | str):
        self.parameters = ParamMap(parameter=parameter)
        self.generator = I(target) - Z(target)
        super().__init__((target,))

    @classmethod
    def num_parameters(cls) -> int:
        return 1

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([0, 2], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        lmbda = torch.exp(1j * evaluate(self.parameters.parameter, as_torch=True))
        return torch.cat((torch.ones(1), lmbda))


class I(PrimitiveBlock):
    """The identity gate"""

    name = OpName.I

    def __init__(self, target: int):
        super().__init__((target,))

    def __ixor__(self, other: AbstractBlock | int) -> AbstractBlock:
        if not isinstance(other, AbstractBlock):
            raise ValueError(
                f"Can only initialize a kron block with another block. Got {type(other)}."
            )
        return other

    def __imul__(self, other: AbstractBlock | TNumber | Parameter) -> AbstractBlock:
        if not isinstance(other, AbstractBlock):
            raise ValueError(
                "In-place multiplication is available " "only for AbstractBlock instances"
            )
        return other

    @property
    def generator(self) -> AbstractBlock:
        return I(*self.qubit_support)

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.ones(2, dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.ones(2, dtype=cdouble)

    def __ascii__(self, console: Console) -> Padding:
        return Padding("──────", (1, 1, 1, 1))

    def dagger(self) -> I:
        return I(*self.qubit_support)


TPauliBlock = Union[X, Y, Z, I, N]


class H(PrimitiveBlock):
    """The Hadamard or H gate"""

    name = OpName.H

    def __init__(self, target: int):
        self.generator = (1 / np.sqrt(2)) * (X(target) + Z(target) - np.sqrt(2) * I(target))
        super().__init__((target,))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([-2, 0], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.tensor([-1, 1], dtype=cdouble)

    def dagger(self) -> H:
        return H(*self.qubit_support)


class Zero(PrimitiveBlock):
    name = OpName.ZERO

    def __init__(self) -> None:
        self.generator = 0 * I(0)
        super().__init__((0,))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.zeros(2, dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.zeros(2, dtype=cdouble)

    def __add__(self, other: AbstractBlock) -> AbstractBlock:
        return other

    def __iadd__(self, other: AbstractBlock) -> AbstractBlock:
        return other

    def __sub__(self, other: AbstractBlock) -> AbstractBlock:
        return -other

    def __isub__(self, other: AbstractBlock) -> AbstractBlock:
        return -other

    def __mul__(self, other: AbstractBlock | TNumber | Parameter) -> AbstractBlock:
        return self

    def __imul__(self, other: AbstractBlock | TNumber | Parameter) -> AbstractBlock:
        return self

    def __ixor__(self, other: AbstractBlock | TNumber | Parameter) -> AbstractBlock:
        return self

    def __pow__(self, power: int) -> AbstractBlock:
        return self

    def dagger(self) -> Zero:
        return Zero()


class RX(ParametricBlock):
    """The Rx gate"""

    name = OpName.RX

    def __init__(self, target: int, parameter: Parameter | TParameter | ParamMap):
        # TODO: should we give them more meaningful names? like 'angle'?
        self.parameters = (
            parameter if isinstance(parameter, ParamMap) else ParamMap(parameter=parameter)
        )
        self.generator = X(target)
        super().__init__((target,))

    @classmethod
    def num_parameters(cls) -> int:
        return 1

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([-1, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        val = evaluate(self.parameters.parameter, as_torch=True)
        lmbd = torch.cos(val / 2.0) - 1j * torch.sin(val / 2.0)
        return torch.cat((lmbd, lmbd.conj()))


class RY(ParametricBlock):
    """The Ry gate"""

    name = OpName.RY

    def __init__(self, target: int, parameter: Parameter | TParameter | ParamMap):
        self.parameters = (
            parameter if isinstance(parameter, ParamMap) else ParamMap(parameter=parameter)
        )
        self.generator = Y(target)
        super().__init__((target,))

    @classmethod
    def num_parameters(cls) -> int:
        return 1

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([-1, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        val = evaluate(self.parameters.parameter, as_torch=True)
        lmbd = torch.cos(val / 2.0) - 1j * torch.sin(val / 2.0)
        return torch.cat((lmbd, lmbd.conj()))


class RZ(ParametricBlock):
    """The Rz gate"""

    name = OpName.RZ

    def __init__(self, target: int, parameter: Parameter | TParameter | ParamMap):
        self.parameters = (
            parameter if isinstance(parameter, ParamMap) else ParamMap(parameter=parameter)
        )
        self.generator = Z(target)
        super().__init__((target,))

    @classmethod
    def num_parameters(cls) -> int:
        return 1

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([-1, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        val = evaluate(self.parameters.parameter, as_torch=True)
        lmbd = torch.cos(val / 2.0) - 1j * torch.sin(val / 2.0)
        return torch.cat((lmbd, lmbd.conj()))


class U(ParametricBlock):
    """Arbitrary one-qubit rotation in the Bloch sphere

    This operation accepts 3 parameters (phi, theta, omega)"""

    name = OpName.U

    def __init__(
        self,
        target: int,
        phi: Parameter | TParameter,
        theta: Parameter | TParameter,
        omega: Parameter | TParameter,
    ):
        self.parameters = ParamMap(phi=phi, theta=theta, omega=omega)
        self.generator = chain(Z(target), Y(target), Z(target))
        super().__init__((target,))

    @classmethod
    def num_parameters(cls) -> int:
        return 3

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([-1, 1], dtype=torch.cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        phi = evaluate(self.parameters.phi)
        theta = evaluate(self.parameters.theta)
        omega = evaluate(self.parameters.omega)
        lmbd = np.exp(-1j * (phi + omega) / 2) * np.cos(theta / 2)
        return torch.cat((lmbd, lmbd.conj()))

    @property
    def n_qubits(self) -> int:
        return 1

    def digital_decomposition(self) -> AbstractBlock:
        return chain(
            RZ(self.qubit_support[0], self.parameters.phi),
            RY(self.qubit_support[0], self.parameters.theta),
            RZ(self.qubit_support[0], self.parameters.omega),
        )


class HamEvo(TimeEvolutionBlock):
    """
    A block implementing the Hamiltonian evolution operation H where:
        H = exp(-iG, t)
    where G represents a square generator and t represents the time parameter
    which can be parametrized.

    Arguments:
        generator: Either a AbstractBlock, torch.Tensor or numpy.ndarray.
        parameter: A scalar or vector of numeric or torch.Tensor type.
        qubit_support: The qubits on which the evolution will be performed on.

    Examples:

    ```python exec="on" source="material-block" result="json"
    from qadence import RX, HamEvo, run
    import torch
    hevo = HamEvo(generator=RX(0, torch.pi), parameter=torch.rand(2))
    print(run(hevo))
    # Now lets use a torch.Tensor as a generator, Now we have to pass the support
    gen = torch.rand(2,2, dtype=torch.complex128)
    hevo = HamEvo(generator=gen, parameter=torch.rand(2), qubit_support=(0,))
    print(run(hevo))
    ```
    """

    name = OpName.HAMEVO
    draw_generator: bool = False

    def __init__(
        self,
        generator: Union[TGenerator, AbstractBlock],
        parameter: TParameter,
        qubit_support: tuple[int, ...] = None,
    ):
        gen_exprs = {}
        if qubit_support is None and not isinstance(generator, AbstractBlock):
            raise ValueError("You have to supply a qubit support for non-block generators.")
        super().__init__(qubit_support if qubit_support else generator.qubit_support)
        if isinstance(generator, AbstractBlock):
            qubit_support = generator.qubit_support
            if generator.is_parametric:
                gen_exprs = {str(e): e for e in expressions(generator)}
        elif isinstance(generator, torch.Tensor):
            msg = "Please provide a square generator."
            if len(generator.shape) == 2:
                assert generator.shape[0] == generator.shape[1], msg
            elif len(generator.shape) == 3:
                assert generator.shape[1] == generator.shape[2], msg
                assert generator.shape[0] == 1, "Qadence doesnt support batched generators."
            else:
                raise TypeError(
                    "Only 2D or 3D generators are supported.\
                                In case of a 3D generator, the batch dim\
                                is expected to be at dim 0."
                )
            gen_exprs = {str(generator.__hash__()): generator}
        elif isinstance(generator, (sympy.Basic, sympy.Array)):
            gen_exprs = {str(generator): generator}
        else:
            raise TypeError(
                f"Generator of type {type(generator)} not supported.\
                            If you're using a numpy.ndarray, please cast it to a torch tensor."
            )
        ps = {"parameter": Parameter(parameter), **gen_exprs}
        self.parameters = ParamMap(**ps)
        self.generator = generator

    @classmethod
    def num_parameters(cls) -> int:
        return 2

    @cached_property
    def eigenvalues_generator(
        self, max_num_evals: int | None = None, max_num_gaps: int | None = None
    ) -> Tensor:
        if isinstance(self.generator, AbstractBlock):
            generator_tensor = block_to_tensor(self.generator)
        elif isinstance(self.generator, Tensor):
            generator_tensor = self.generator
        return eigenvalues(generator_tensor, max_num_evals, max_num_gaps)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.exp(
            -1j * evaluate(self.parameters.parameter, as_torch=True) * self.eigenvalues_generator
        )

    @property
    def n_qubits(self) -> int:
        if isinstance(self.generator, Tensor):
            n_qubits = int(np.log2(self.generator.shape[-1]))
        else:
            n_qubits = self.generator.n_qubits  # type: ignore [union-attr]

        return n_qubits

    def dagger(self) -> Any:
        p = list(self.parameters.expressions())[0]
        return HamEvo(deepcopy(self.generator), -extract_original_param_entry(p))

    def digital_decomposition(self, approximation: LTSOrder = LTSOrder.ST4) -> AbstractBlock:
        """Decompose the Hamiltonian evolution into digital gates

        Args:
            approximation (str, optional): Choose the type of decomposition. Defaults to "st4".
                Available types are:
                * 'basic' = apply first-order Trotter formula and decompose each term of
                    the exponential into digital gates. It is exact only if applied to an
                    operator whose terms are mutually commuting.
                * 'st2' = Trotter-Suzuki 2nd order formula for approximating non-commuting
                    Hamiltonians.
                * 'st4' = Trotter-Suzuki 4th order formula for approximating non-commuting
                    Hamiltonians.

        Returns:
            AbstractBlock: a block with the digital decomposition
        """

        # psi(t) = exp(-i * H * t * psi0)
        # psi(t) = exp(-i * lambda * t * psi0)
        # H = sum(Paulin) + sum(Pauli1*Pauli2)
        logger.info("Quantum simulation of the time-independent Schrödinger equation.")

        blocks = []

        # how to change the type/dict to enum effectively

        # when there is a term including non-commuting matrices use st2 or st4

        # 1) should check that the given generator respects the constraints
        # single-qubit gates

        assert isinstance(
            self.generator, AbstractBlock
        ), "Only a generator represented as a block can be decomposed"

        if block_is_qubit_hamiltonian(self.generator):
            try:
                block_is_commuting_hamiltonian(self.generator)
                approximation = LTSOrder.BASIC  # use the simpler approach if the H is commuting
            except TypeError:
                logger.warning(
                    """Non-commuting terms in the Pauli operator.
                    The Suzuki-Trotter approximation is applied."""
                )

            blocks.extend(
                lie_trotter_suzuki(
                    block=self.generator,
                    parameter=self.parameters.parameter,
                    order=LTSOrder[approximation],
                )
            )

            # 2) return an AbstractBlock instance with the set of gates
            # resulting from the decomposition

            return chain(*blocks)
        else:
            raise NotImplementedError(
                "The current digital decomposition can be applied only to Pauli Hamiltonians."
            )


class CNOT(ControlBlock):
    """The CNot, or CX, gate"""

    name = OpName.CNOT

    def __init__(self, control: int, target: int) -> None:
        self.generator = kron((I(control) - Z(control)) * 0.5, X(target) - I(target))
        super().__init__((control,), X(target))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([-2, 0, 0, 0], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.tensor([-1, 1, 1, 1], dtype=cdouble)

    def __ascii__(self, console: Console) -> RenderableType:
        (target, control) = self.qubit_support
        h = abs(target - control) + 1
        return Panel(self._block_title, expand=False, height=3 * h)

    def __rich_tree__(self, tree: Tree = None) -> Tree:
        if tree is None:
            return Tree(self._block_title)
        else:
            tree.add(self._block_title)
        return tree

    def dagger(self) -> CNOT:
        return CNOT(*self.qubit_support)


class MCZ(ControlBlock):
    name = OpName.MCZ

    def __init__(self, control: tuple[int, ...], target: int) -> None:
        self.generator = kron(
            *[(I(qubit) - Z(qubit)) * 0.5 for qubit in control], Z(target) - I(target)
        )
        super().__init__(control, Z(target))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.cat((torch.tensor(-2, dtype=cdouble), torch.zeros(2**self.n_qubits - 1)))

    @property
    def eigenvalues(self) -> Tensor:
        torch.cat((torch.tensor(-1, dtype=cdouble), torch.ones(2**self.n_qubits - 1)))

    def __ascii__(self, console: Console) -> RenderableType:
        (target, control) = self.qubit_support
        h = abs(target - control) + 1
        return Panel(self._block_title, expand=False, height=3 * h)

    def __rich_tree__(self, tree: Tree = None) -> Tree:
        if tree is None:
            return Tree(self._block_title)
        else:
            tree.add(self._block_title)
        return tree

    def dagger(self) -> MCZ:
        return MCZ(self.qubit_support[:-1], self.qubit_support[-1])


class CZ(MCZ):
    """The CZ gate"""

    name = OpName.CZ

    def __init__(self, control: int, target: int) -> None:
        super().__init__((control,), target)

    def dagger(self) -> CZ:
        return CZ(self.qubit_support[-2], self.qubit_support[-1])


class MCRX(ParametricControlBlock):
    name = OpName.MCRX

    def __init__(
        self,
        control: tuple[int, ...],
        target: int,
        parameter: Parameter | TNumber | sympy.Expr | str,
    ) -> None:
        self.generator = kron(*[(I(qubit) - Z(qubit)) * 0.5 for qubit in control], X(target))
        super().__init__(control, RX(target, parameter))

    @classmethod
    def num_parameters(cls) -> int:
        return 1

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.cat(
            (torch.zeros(2**self.n_qubits - 2), torch.tensor([1, -1], dtype=cdouble))
        )

    @property
    def eigenvalues(self) -> Tensor:
        val = evaluate(self.parameters.parameter, as_torch=True)
        lmbd = torch.cos(val / 2.0) - 1j * torch.sin(val / 2.0)
        return torch.cat((torch.ones(2**self.n_qubits - 2), lmbd, lmbd.conj()))


class CRX(MCRX):
    """The CRX gate"""

    name = OpName.CRX

    def __init__(
        self,
        control: int,
        target: int,
        parameter: Parameter | TNumber | sympy.Expr | str,
    ):
        super().__init__((control,), target, parameter)


class MCRY(ParametricControlBlock):
    name = OpName.MCRY

    def __init__(
        self,
        control: tuple[int, ...],
        target: int,
        parameter: Parameter | TNumber | sympy.Expr | str,
    ) -> None:
        self.generator = kron(*[(I(qubit) - Z(qubit)) * 0.5 for qubit in control], Y(target))
        super().__init__(control, RY(target, parameter))

    @classmethod
    def num_parameters(cls) -> int:
        return 1

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.cat(
            (torch.zeros(2**self.n_qubits - 2), torch.tensor([1, -1], dtype=cdouble))
        )

    @property
    def eigenvalues(self) -> Tensor:
        val = evaluate(self.parameters.parameter, as_torch=True)
        lmbd = torch.cos(val / 2.0) - 1j * torch.sin(val / 2.0)
        return torch.cat((torch.ones(2**self.n_qubits - 2), lmbd, lmbd.conj()))


class CRY(MCRY):
    """The CRY gate"""

    name = OpName.CRY

    def __init__(
        self,
        control: int,
        target: int,
        parameter: Parameter | TNumber | sympy.Expr | str,
    ):
        super().__init__((control,), target, parameter)


class MCRZ(ParametricControlBlock):
    name = OpName.MCRZ

    def __init__(
        self,
        control: tuple[int, ...],
        target: int,
        parameter: Parameter | TNumber | sympy.Expr | str,
    ) -> None:
        self.generator = kron(*[(I(qubit) - Z(qubit)) * 0.5 for qubit in control], Z(target))
        super().__init__(control, RZ(target, parameter))

    @classmethod
    def num_parameters(cls) -> int:
        return 1

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.cat(
            (torch.zeros(2**self.n_qubits - 2), torch.tensor([1, -1], dtype=cdouble))
        )

    @property
    def eigenvalues(self) -> Tensor:
        val = evaluate(self.parameters.parameter, as_torch=True)
        lmbd = torch.cos(val / 2.0) - 1j * torch.sin(val / 2.0)
        return torch.cat((torch.ones(2**self.n_qubits - 2), lmbd, lmbd.conj()))


class CRZ(MCRZ):
    """The CRZ gate"""

    name = OpName.CRZ

    def __init__(
        self,
        control: int,
        target: int,
        parameter: Parameter | TNumber | sympy.Expr | str,
    ):
        super().__init__((control,), target, parameter)


class CSWAP(ControlBlock):
    """The CSWAP (Control-SWAP) gate."""

    name = OpName.CSWAP

    def __init__(self, control: int | tuple[int, ...], target1: int, target2: int) -> None:
        if isinstance(control, tuple):
            control = control[0]

        a00m = 0.5 * (Z(control) - I(control))
        a00p = -0.5 * (Z(control) + I(control))
        a11 = 0.5 * (Z(target1) - I(target1))
        a22 = -0.5 * (Z(target2) + I(target2))
        a12 = 0.5 * (chain(X(target1), Z(target1)) + X(target1))
        a21 = 0.5 * (chain(Z(target2), X(target2)) + X(target2))
        no_effect = kron(a00m, I(target1), I(target2))
        swap_effect = (
            kron(a00p, -1.0 * a22, a11)
            + kron(a00p, -1.0 * a11, a22)
            + kron(a00p, a12, a21)
            + kron(a00p, a21, a12)
        )
        self.generator = no_effect + swap_effect
        super().__init__((control,), SWAP(target1, target2))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor((1, -1, 1, 1, 1, 1, 1, 1), dtype=torch.cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.tensor((1, -1, 1, 1, 1, 1, 1, 1), dtype=torch.cdouble)

    @property
    def nqubits(self) -> int:
        return 3

    def dagger(self) -> CSWAP:
        return CSWAP(*self.qubit_support)


class T(PrimitiveBlock):
    """The T gate"""

    name = OpName.T

    def __init__(self, target: int):
        self.generator = I(target) - Z(target)
        super().__init__((target,))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([0, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.tensor([1.0, torch.sqrt(torch.tensor([1j]))], dtype=cdouble)

    @property
    def n_qubits(self) -> int:
        return 1

    def dagger(self) -> TDagger:
        return TDagger(*self.qubit_support)


class TDagger(PrimitiveBlock):
    """The Hermitian adjoint/conjugate transpose of the T gate"""

    # FIXME: this gate is not support by any backend
    name = "T_dagger"

    def __init__(self, target: int):
        self.generator = I(target) - Z(target)
        super().__init__((target,))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([0, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.tensor([1.0, torch.sqrt(torch.tensor([-1j]))], dtype=cdouble)

    @property
    def n_qubits(self) -> int:
        return 1

    def dagger(self) -> T:
        return T(*self.qubit_support)


class SWAP(PrimitiveBlock):
    """The SWAP gate"""

    name = OpName.SWAP

    def __init__(self, control: int, target: int) -> None:
        a11 = 0.5 * (Z(control) - I(control))
        a22 = -0.5 * (Z(target) + I(target))
        a12 = 0.5 * (chain(X(control), Z(control)) + X(control))
        a21 = 0.5 * (chain(Z(target), X(target)) + X(target))
        self.generator = (
            kron(-1.0 * a22, a11) + kron(-1.0 * a11, a22) + kron(a12, a21) + kron(a21, a12)
        )
        super().__init__((control, target))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([-2, 0, 0, 0], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.tensor([-1, 1, 1, 1], dtype=cdouble)

    @property
    def n_qubits(self) -> int:
        return 2

    @property
    def _block_title(self) -> str:
        c, t = self.qubit_support
        s = f"{self.name}({c}, {t})"
        return s if self.tag is None else (s + rf" \[tag: {self.tag}]")

    def dagger(self) -> SWAP:
        return SWAP(*self.qubit_support)


class AnalogSWAP(HamEvo):
    """
    Single time-independent Hamiltonian evolution over a Rydberg Ising
    hamiltonian yielding a SWAP (up to global phase).

    Derived from
    [Bapat et al.](https://journals.aps.org/prresearch/abstract/10.1103/PhysRevResearch.4.L012023)
    where it is applied to XX-type Hamiltonian
    """

    name = OpName.ANALOGSWAP

    def __init__(self, control: int, target: int, parameter: TParameter = 3 * np.pi / 4):
        rydberg_ising_hamiltonian_generator = (
            4.0 * kron((I(control) - Z(control)) / 2.0, (I(target) - Z(target)) / 2.0)
            + (2.0 / 3.0) * np.sqrt(2.0) * X(control)
            + (2.0 / 3.0) * np.sqrt(2.0) * X(target)
            + (1.0 + np.sqrt(5.0) / 3) * Z(control)
            + (1.0 + np.sqrt(5.0) / 3) * Z(target)
        )
        super().__init__(rydberg_ising_hamiltonian_generator, parameter, (control, target))


class MCPHASE(ParametricControlBlock):
    name = OpName.MCPHASE

    def __init__(
        self,
        control: tuple[int, ...],
        target: int,
        parameter: Parameter | TNumber | sympy.Expr | str,
    ) -> None:
        self.generator = kron(
            *[(I(qubit) - Z(qubit)) * 0.5 for qubit in control], Z(target) - I(target)
        )
        super().__init__(control, PHASE(target, parameter))

    @classmethod
    def num_parameters(cls) -> int:
        return 1

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.cat(
            (torch.tensor([-2, 0], dtype=cdouble), (torch.zeros(2**self.n_qubits - 2)))
        )

    @property
    def eigenvalues(self) -> Tensor:
        v = evaluate(self.parameters.parameter, as_torch=True)
        return torch.cat((torch.ones(2**self.n_qubits - 1), torch.exp(1j * v)))

    def __rich_tree__(self, tree: Tree = None) -> Tree:
        if tree is None:
            return Tree(self._block_title)
        else:
            tree.add(self._block_title)
        return tree

    def __ascii__(self, console: Console) -> RenderableType:
        (target, control) = self.qubit_support
        h = abs(target - control) + 1
        return Panel(self._block_title, expand=False, height=3 * h)


class CPHASE(MCPHASE):
    """The CPHASE gate"""

    name = OpName.CPHASE

    def __init__(
        self,
        control: int,
        target: int,
        parameter: Parameter | TNumber | sympy.Expr | str,
    ):
        super().__init__((control,), target, parameter)


class Toffoli(ControlBlock):
    name = OpName.TOFFOLI

    def __init__(self, control: tuple[int, ...], target: int) -> None:
        self.generator = kron(
            *[(I(qubit) - Z(qubit)) * 0.5 for qubit in control], X(target) - I(target)
        )
        super().__init__(control, X(target))

    def dagger(self) -> Toffoli:
        return Toffoli(self.qubit_support[:-1], self.qubit_support[-1])

    @property
    def n_qubits(self) -> int:
        return len(self.qubit_support)

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor(
            [-2, *[0 for _ in range(2 ** len(self.qubit_support) - 1)]], dtype=cdouble
        )

    @property
    def eigenvalues(self) -> Tensor:
        return torch.tensor(
            [-1, *[1 for _ in range(2 ** len(self.qubit_support) - 1)]], dtype=cdouble
        )


# FIXME: better name that stresses difference to `Wait`?
@dataclass(eq=False, repr=False)
class AnalogEntanglement(AnalogBlock):
    parameters: ParamMap = ParamMap(duration=1.0)
    qubit_support: QubitSupport = QubitSupport("global")

    @property
    def eigenvalues_generator(self) -> torch.Tensor:
        return torch.tensor([0.0], dtype=cdouble)

    @property
    def duration(self) -> Basic:
        return self.parameters.duration


def _cast(T: Any, val: Any) -> Any:
    return val if isinstance(val, T) else T(val)


def wait(
    duration: TNumber | sympy.Basic,
    qubit_support: str | QubitSupport | tuple = "global",
) -> WaitBlock:
    """Constructs a [`WaitBlock`][qadence.blocks.analog.WaitBlock].

    Arguments:
        duration: Time to wait in nanoseconds.
        qubit_support: Qubits the `WaitBlock` is applied to. Can be either
            `"global"` to apply the wait block to all qubits or a tuple of integers.

    Returns:
        a `WaitBlock`
    """
    q = _cast(QubitSupport, qubit_support)
    ps = ParamMap(duration=duration)
    return WaitBlock(parameters=ps, qubit_support=q)


def entangle(
    duration: Any,
    qubit_support: str | QubitSupport | Tuple = "global",
) -> AnalogEntanglement:
    q = _cast(QubitSupport, qubit_support)
    ps = ParamMap(duration=duration)
    return AnalogEntanglement(parameters=ps, qubit_support=q)


def AnalogRot(
    duration: float | str | Parameter = 1000.0,
    omega: float | str | Parameter = 0,
    delta: float | str | Parameter = 0,
    phase: float | str | Parameter = 0,
    qubit_support: str | QubitSupport | Tuple = "global",
) -> ConstantAnalogRotation:
    """General analog rotation operation.

    Arguments:
        duration: Duration of the rotation [ns].
        omega: Rotation frequency [rad/μs]
        delta: Rotation frequency [rad/μs]
        phase: Phase angle [rad]
        qubit_support: Defines the (local/global) qubit support

    Returns:
        ConstantAnalogRotation
    """
    q = _cast(QubitSupport, qubit_support)
    if isinstance(duration, str):
        duration = Parameter(duration)
    alpha = duration * sympy.sqrt(omega**2 + delta**2) / 1000  # type: ignore [operator]

    ps = ParamMap(alpha=alpha, duration=duration, omega=omega, delta=delta, phase=phase)
    return ConstantAnalogRotation(parameters=ps, qubit_support=q)


def _analog_rot(
    angle: float | str | Parameter,
    qubit_support: str | QubitSupport | Tuple,
    phase: float,
) -> ConstantAnalogRotation:
    q = _cast(QubitSupport, qubit_support)
    # assuming some arbitrary omega = π rad/μs
    alpha = _cast(Parameter, angle)

    omega = np.pi
    duration = alpha / omega * 1000

    # FIXME: once https://gitlab.pasqal.com/pqs/qadence/-/issues/402 is fixed set default duration
    # in the function arguments to:
    # duration = Parameter(160)
    # and compute omega like this:
    # omega = alpha / duration * 1000
    ps = ParamMap(alpha=alpha, duration=duration, omega=omega, delta=0, phase=phase)
    return ConstantAnalogRotation(parameters=ps, qubit_support=q)


def AnalogRX(
    angle: float | str | Parameter,
    qubit_support: str | QubitSupport | Tuple = "global",
) -> ConstantAnalogRotation:
    """Analog X rotation. Shorthand for [`AnalogRot`][qadence.operations.AnalogRot]:

    ```python
    φ=2.4; Ω=π; t = φ/Ω * 1000
    AnalogRot(duration=t, omega=Ω)
    ```

    Arguments:
        angle: Rotation angle [rad]
        qubit_support: Defines the (local/global) qubit support

    Returns:
        ConstantAnalogRotation
    """
    return _analog_rot(angle, qubit_support, phase=0)


def AnalogRY(
    angle: float | str | Parameter,
    qubit_support: str | QubitSupport | Tuple = "global",
) -> ConstantAnalogRotation:
    """Analog Y rotation. Shorthand for [`AnalogRot`][qadence.operations.AnalogRot]:

    ```python
    φ=2.4; Ω=π; t = φ/Ω * 1000
    AnalogRot(duration=t, omega=Ω, phase=-π/2)
    ```
    Arguments:
        angle: Rotation angle [rad]
        qubit_support: Defines the (local/global) qubit support

    Returns:
        ConstantAnalogRotation
    """
    return _analog_rot(angle, qubit_support, phase=-np.pi / 2)


def AnalogRZ(
    angle: float | str | Parameter,
    qubit_support: str | QubitSupport | Tuple = "global",
) -> ConstantAnalogRotation:
    """Analog Z rotation. Shorthand for [`AnalogRot`][qadence.operations.AnalogRot]:
    ```
    φ=2.4; δ=π; t = φ/δ * 100)
    AnalogRot(duration=t, delta=δ, phase=π/2)
    ```
    """
    q = _cast(QubitSupport, qubit_support)
    alpha = _cast(Parameter, angle)
    delta = np.pi
    duration = alpha / delta * 1000
    ps = ParamMap(alpha=alpha, duration=duration, omega=0, delta=delta, phase=np.pi / 2)
    return ConstantAnalogRotation(qubit_support=q, parameters=ps)


# gate sets
# FIXME: this could be inferred by the number of qubits if we had
# a class property for each operation. The number of qubits can default
# to None for operations which do not have it by default
# this would allow to greatly simplify the tests
pauli_gateset: list = [I, X, Y, Z]
# FIXME: add Tdagger when implemented
single_qubit_gateset = [X, Y, Z, H, I, RX, RY, RZ, U, S, SDagger, T, PHASE]
two_qubit_gateset = [CNOT, SWAP, CZ, CRX, CRY, CRZ, CPHASE]
three_qubit_gateset = [CSWAP]
multi_qubit_gateset = [Toffoli, MCRX, MCRY, MCRZ, MCPHASE, MCZ]
analog_gateset = [
    HamEvo,
    ConstantAnalogRotation,
    AnalogEntanglement,
    AnalogSWAP,
    AnalogRX,
    AnalogRY,
    AnalogRZ,
    entangle,
    wait,
]
non_unitary_gateset = [Zero, N]



from __future__ import annotations

from functools import reduce
from itertools import chain as flatten
from operator import add
from typing import Callable, Sequence

import pyqtorch.modules as pyq
import sympy
import torch
from pyqtorch.core.utils import _apply_batch_gate
from torch.nn import Module
from torch.utils.checkpoint import checkpoint

from qadence.blocks import (
    AbstractBlock,
    AddBlock,
    ChainBlock,
    CompositeBlock,
    MatrixBlock,
    ParametricBlock,
    PrimitiveBlock,
    ScaleBlock,
    TimeEvolutionBlock,
)
from qadence.blocks.block_to_tensor import (
    _block_to_tensor_embedded,
    block_to_diagonal,
    block_to_tensor,
)
from qadence.operations import OpName, U

from .config import Configuration

# Tdagger is not supported currently
supported_gates = list(set(OpName.list()) - set([OpName.TDAGGER]))
"""The set of supported gates. Tdagger is currently not supported."""


def is_single_qubit_chain(block: AbstractBlock) -> bool:
    return (
        isinstance(block, (ChainBlock))
        and block.n_supports == 1
        and all([isinstance(b, (ParametricBlock, PrimitiveBlock)) for b in block])
        and not any([isinstance(b, (ScaleBlock, U)) for b in block])
    )


def convert_observable(
    block: AbstractBlock, n_qubits: int, config: Configuration = None
) -> Sequence[Module]:
    return [PyQObservable(block, n_qubits, config)]


def convert_block(
    block: AbstractBlock, n_qubits: int = None, config: Configuration = None
) -> Sequence[Module]:
    if n_qubits is None:
        n_qubits = max(block.qubit_support) + 1

    if config is None:
        config = Configuration()

    if isinstance(block, ScaleBlock):
        return [ScalePyQOperation(n_qubits, block, config)]

    elif isinstance(block, AddBlock):
        ops = list(flatten(*(convert_block(b, n_qubits, config) for b in block.blocks)))
        return [AddPyQOperation(block.qubit_support, n_qubits, ops, config)]

    elif isinstance(block, ParametricBlock):
        if isinstance(block, TimeEvolutionBlock):
            op = HEvoPyQOperation(
                qubits=block.qubit_support,
                n_qubits=n_qubits,
                # TODO: use the hevo_algo configuration here to switch between different algorithms
                # for executing the Hamiltonian evolution
                operation=pyq.HamiltonianEvolution(
                    block.qubit_support,
                    n_qubits,
                    n_steps=config.n_steps_hevo,
                ),
                block=block,
                config=config,
            )
        else:
            op = ParametricPyQOperation(n_qubits, block, config)
        return [op]
    elif isinstance(block, MatrixBlock):
        return [PyQMatrixBlock(block, n_qubits, config)]
    elif isinstance(block, PrimitiveBlock):
        return [PyQOperation(n_qubits, block)]
    elif isinstance(block, CompositeBlock):
        ops = list(flatten(*(convert_block(b, n_qubits, config) for b in block.blocks)))
        if is_single_qubit_chain(block) and config.use_single_qubit_composition:
            return [PyQComposedBlock(ops, block.qubit_support, n_qubits, config)]
        else:
            # NOTE: without wrapping in a pyq.QuantumCircuit here the kron/chain
            # blocks won't be properly nested which leads to incorrect results from
            # the `AddBlock`s. For example:
            # add(chain(Z(0), Z(1))) has to result in the following (pseudo-code)
            # AddPyQOperation(pyq.QuantumCircuit(Z, Z))
            # as opposed to
            # AddPyQOperation(Z, Z)
            # which would be wrong.
            return [pyq.QuantumCircuit(n_qubits, ops)]

    else:
        msg = (
            f"Non supported operation of type {type(block)}. "
            "In case you are trying to run an `AnalogBlock`, try converting it "
            "with `add_interaction` first."
        )
        raise NotImplementedError(msg)


class PyQOperation(Module):
    def __init__(self, n_qubits: int, block: AbstractBlock):
        super().__init__()
        name = block.name[1:] if block.name.startswith("MC") else block.name
        Op = getattr(pyq, name)
        self.operation = Op(block.qubit_support, n_qubits)

    # primitive blocks do not require any parameter value, hence the
    # second empty argument added here
    def forward(self, state: torch.Tensor, _: dict[str, torch.Tensor] = None) -> torch.Tensor:
        return self.apply(self.matrices(), state)

    def matrices(self, _: dict[str, torch.Tensor] = None) -> torch.Tensor:
        return self.operation.matrix

    def apply(self, matrices: torch.Tensor, state: torch.Tensor) -> torch.Tensor:
        return self.operation.apply(matrices, state)


class ParametricPyQOperation(Module):
    def __init__(self, n_qubits: int, block: ParametricBlock, config: Configuration):
        super().__init__()
        name = block.name[1:] if block.name.startswith("MC") else block.name
        Op = getattr(pyq, name)
        self.operation = Op(block.qubit_support, n_qubits)
        self.param_names = config.get_param_name(block)
        num_params = len(self.param_names)
        if num_params == 1:

            def _fwd(state: torch.Tensor, values: dict[str, torch.Tensor]) -> torch.Tensor:
                return self.apply(self.matrices(values), state)

        else:

            def _fwd(state: torch.Tensor, values: dict[str, torch.Tensor]) -> torch.Tensor:
                op_params = {key: values[key] for key in self.param_names}
                max_batch_size = max(p.size() for p in values.values())
                new_values = {
                    k: (v if v.size() == max_batch_size else v.repeat(max_batch_size, 1, 1))
                    for k, v in op_params.items()
                }
                return self.apply(self.matrices(new_values), state)

        if config.use_gradient_checkpointing:

            def _forward(state: torch.Tensor, values: dict[str, torch.Tensor]) -> torch.Tensor:
                return checkpoint(_fwd, state, values, use_reentrant=False)

        else:

            def _forward(state: torch.Tensor, values: dict[str, torch.Tensor]) -> torch.Tensor:
                return _fwd(state, values)

        self._forward = _forward

    def matrices(self, values: dict[str, torch.Tensor]) -> torch.Tensor:
        thetas = torch.vstack([values[name] for name in self.param_names])
        return self.operation.matrices(thetas)

    def apply(self, matrices: torch.Tensor, state: torch.Tensor) -> torch.Tensor:
        return self.operation.apply(matrices, state)

    def forward(self, state: torch.Tensor, values: dict[str, torch.Tensor]) -> torch.Tensor:
        return self._forward(state, values)


class PyQMatrixBlock(Module):
    def __init__(self, block: MatrixBlock, n_qubits: int, config: Configuration = None):
        super().__init__()
        self.n_qubits = n_qubits
        self.qubits = block.qubit_support
        self.register_buffer("mat", block.matrix.unsqueeze(2))

    def forward(self, state: torch.Tensor, _: dict[str, torch.Tensor] = None) -> torch.Tensor:
        return self.apply(self.mat, state)

    def apply(self, matrices: torch.Tensor, state: torch.Tensor) -> torch.Tensor:
        batch_size = state.size(-1)
        return _apply_batch_gate(state, matrices, self.qubits, self.n_qubits, batch_size)


class PyQComposedBlock(Module):
    def __init__(
        self,
        ops: list[Module],
        qubits: list[int] | tuple,
        n_qubits: int,
        config: Configuration = None,
    ):
        """Compose a chain of single qubit operations on the same qubit into a single
        call to _apply_batch_gate."""
        super().__init__()
        self.operations = ops
        self.qubits = qubits
        self.n_qubits = n_qubits

    def forward(
        self, state: torch.Tensor, values: dict[str, torch.Tensor] | None = None
    ) -> torch.Tensor:
        batch_size = state.size(-1)
        return self.apply(self.matrices(values, batch_size), state)

    def apply(self, matrices: torch.Tensor, state: torch.Tensor) -> torch.Tensor:
        batch_size = state.size(-1)
        return _apply_batch_gate(state, matrices, self.qubits, self.n_qubits, batch_size)

    def matrices(self, values: dict[str, torch.Tensor] | None, batch_size: int) -> torch.Tensor:
        perm = (2, 0, 1)  # We permute the dims since torch.bmm expects the batch_dim at 0.

        def _expand_mat(m: torch.Tensor) -> torch.Tensor:
            if len(m.size()) == 2:
                m = m.unsqueeze(2).repeat(
                    1, 1, batch_size
                )  # Primitive gates are 2D, so we expand them.
            elif m.shape != (2, 2, batch_size):
                m = m.repeat(1, 1, batch_size)  # In case a tensor is 3D doesnt have batch_size.
            return torch.permute(m, perm)  # This returns shape (batch_size, 2, 2)

        # We reverse the list of tensors here since matmul is not commutative.
        return torch.permute(
            reduce(
                torch.bmm, (_expand_mat(op.matrices(values)) for op in reversed(self.operations))
            ),
            tuple(
                torch.argsort(torch.tensor(perm))
            ),  # We need to undo the permute since PyQ expects (2, 2, batch_size).
        )


class PyQObservable(Module):
    def __init__(self, block: AbstractBlock, n_qubits: int, config: Configuration = None):
        super().__init__()
        if config is None:
            config = Configuration()
        self.n_qubits = n_qubits
        if block._is_diag_pauli and not block.is_parametric:
            diag = block_to_diagonal(block, tuple(range(n_qubits)))
            self.register_buffer("diag", diag)

            def sparse_operation(
                state: torch.Tensor, values: dict[str, torch.Tensor] = None
            ) -> torch.Tensor:
                state = state.reshape(2**self.n_qubits, state.size(-1))
                return (diag * state.T).T

            self.operation = sparse_operation
        else:
            self.operation = pyq.QuantumCircuit(
                n_qubits,
                convert_block(block, n_qubits, config),
            )

        if config.use_gradient_checkpointing:

            def _forward(
                state: torch.Tensor, values: dict[str, torch.Tensor] = None
            ) -> torch.Tensor:
                new_state = checkpoint(self.operation, state, values, use_reentrant=False)
                return pyq.overlap(state, new_state)

        else:

            def _forward(
                state: torch.Tensor, values: dict[str, torch.Tensor] = None
            ) -> torch.Tensor:
                return pyq.overlap(state, self.operation(state, values))

        self._forward = _forward

    def forward(self, state: torch.Tensor, values: dict[str, torch.Tensor]) -> torch.Tensor:
        return self._forward(state, values)


class HEvoPyQOperation(Module):
    def __init__(
        self,
        qubits: Sequence,
        n_qubits: int,
        operation: Callable,
        block: TimeEvolutionBlock,
        config: Configuration,
    ):
        super().__init__()
        self.qubits = qubits
        self.n_qubits = n_qubits
        self.operation = operation
        self.param_names = config.get_param_name(block)
        self._has_parametric_generator: bool
        self.block = block

        if isinstance(block.generator, AbstractBlock) and not block.generator.is_parametric:
            hmat = block_to_tensor(
                block.generator,
                qubit_support=tuple(self.qubits),
                use_full_support=False,
            )
            hmat = hmat.permute(1, 2, 0)

            def _fwd(state: torch.Tensor, values: dict[str, torch.Tensor]) -> torch.Tensor:
                tevo = values[self.param_names[0]]
                return self.operation(hmat, tevo, state)

        elif isinstance(block.generator, torch.Tensor):
            m = block.generator.to(dtype=torch.cdouble)
            hmat = block_to_tensor(
                MatrixBlock(m, qubit_support=block.qubit_support),
                qubit_support=tuple(self.qubits),
                use_full_support=False,
            )
            hmat = hmat.permute(1, 2, 0)

            def _fwd(state: torch.Tensor, values: dict[str, torch.Tensor]) -> torch.Tensor:
                tevo = values[self.param_names[0]]
                return self.operation(hmat, tevo, state)

        elif isinstance(block.generator, sympy.Basic):

            def _fwd(state: torch.Tensor, values: dict[str, torch.Tensor]) -> torch.Tensor:
                tevo = values[self.param_names[0]]
                hmat = values[self.param_names[1]]
                hmat = hmat.squeeze(3)  # FIXME: why is this necessary?
                hmat = hmat.permute(1, 2, 0)
                return self.operation(hmat, tevo, state)

        else:

            def _fwd(state: torch.Tensor, values: dict[str, torch.Tensor]) -> torch.Tensor:
                hmat = _block_to_tensor_embedded(
                    block.generator,  # type: ignore[arg-type]
                    values=values,
                    qubit_support=tuple(self.qubits),
                    use_full_support=False,
                )
                hmat = hmat.permute(1, 2, 0)
                tevo = values[self.param_names[0]]
                return self.operation(hmat, tevo, state)

        if config.use_gradient_checkpointing:

            def _forward(state: torch.Tensor, values: dict[str, torch.Tensor]) -> torch.Tensor:
                return checkpoint(_fwd, state, values, use_reentrant=False)

        else:

            def _forward(state: torch.Tensor, values: dict[str, torch.Tensor]) -> torch.Tensor:
                return _fwd(state, values)

        self._forward = _forward

    def forward(self, state: torch.Tensor, values: dict[str, torch.Tensor]) -> torch.Tensor:
        return self._forward(state, values)


class AddPyQOperation(Module):
    def __init__(
        self, qubits: Sequence, n_qubits: int, operations: list[Module], config: Configuration
    ):
        super().__init__()
        self.operations = operations

        def _fwd(state: torch.Tensor, values: dict[str, torch.Tensor]) -> torch.Tensor:
            return reduce(add, (op(state, values) for op in self.operations))

        if config.use_gradient_checkpointing:

            def _forward(state: torch.Tensor, values: dict[str, torch.Tensor]) -> torch.Tensor:
                return checkpoint(_fwd, state, values, use_reentrant=False)

        else:

            def _forward(state: torch.Tensor, values: dict[str, torch.Tensor]) -> torch.Tensor:
                return _fwd(state, values)

        self._forward = _forward

    def forward(self, state: torch.Tensor, values: dict[str, torch.Tensor]) -> torch.Tensor:
        return self._forward(state, values)


class ScalePyQOperation(Module):
    """
    Computes:

        M = matrix(op, theta)
        scale * matmul(M, state)
    """

    def __init__(self, n_qubits: int, block: ScaleBlock, config: Configuration):
        super().__init__()
        (self.param_name,) = config.get_param_name(block)
        if not isinstance(block.block, PrimitiveBlock):
            raise NotImplementedError(
                "The pyqtorch backend can currently only scale `PrimitiveBlock` types.\
                Please use the following transpile function on your circuit first:\
                from qadence.transpile import scale_primitive_blocks_only"
            )
        self.operation = convert_block(block.block, n_qubits, config)[0]

        def _fwd(state: torch.Tensor, values: dict[str, torch.Tensor]) -> torch.Tensor:
            return values[self.param_name] * self.operation(state, values)

        if config.use_gradient_checkpointing:

            def _forward(state: torch.Tensor, values: dict[str, torch.Tensor]) -> torch.Tensor:
                return checkpoint(_fwd, state, values, use_reentrant=False)

        else:

            def _forward(state: torch.Tensor, values: dict[str, torch.Tensor]) -> torch.Tensor:
                return _fwd(state, values)

        self._forward = _forward

    def matrices(self, values: dict[str, torch.Tensor]) -> torch.Tensor:
        thetas = values[self.param_name]
        return (thetas * self.operation.matrices()).unsqueeze(2)

    def forward(self, state: torch.Tensor, values: dict[str, torch.Tensor]) -> torch.Tensor:
        return self._forward(state, values)



# flake8: noqa F401
from __future__ import annotations

from .api import backend_factory, config_factory
from .pytorch_wrapper import DifferentiableBackend, DiffMode

# Modules to be automatically added to the qadence namespace
__all__ = ["backend_factory", "config_factory", "DifferentiableBackend", "DiffMode"]



from __future__ import annotations

from typing import Callable, Iterable, List

import numpy as np
import sympy
import sympytorch  # type: ignore [import]
import torch
from torch import Tensor

from qadence.blocks import (
    AbstractBlock,
)
from qadence.blocks.utils import (
    expressions,
    parameters,
    uuid_to_expression,
)
from qadence.parameters import evaluate, stringify, torchify

StrTensorDict = dict[str, Tensor]


def unique(x: Iterable) -> List:
    return list(set(x))


def embedding(
    block: AbstractBlock, to_gate_params: bool = False
) -> tuple[StrTensorDict, Callable[[StrTensorDict, StrTensorDict], StrTensorDict],]:
    """Construct embedding function which maps user-facing parameters to either *expression-level*
    parameters or *gate-level* parameters. The construced embedding function has the signature:

         embedding_fn(params: StrTensorDict, inputs: StrTensorDict) -> StrTensorDict:

    which means that it maps the *variational* parameter dict `params` and the *feature* parameter
    dict `inputs` to one new parameter dict `embedded_dict` which holds all parameters that are
    needed to execute a circuit on a given backend. There are two different *modes* for this
    mapping:

    - *Expression-level* parameters: For AD-based optimization. For every unique expression we end
      up with one entry in the embedded dict:
      `len(embedded_dict) == len(unique_parameter_expressions)`.
    - *Gate-level* parameters: For PSR-based optimization or real devices. One parameter for each
      gate parameter, regardless if they are based on the same expression. `len(embedded_dict) ==
      len(parametric_gates)`. This is needed because PSR requires to shift the angles of **every**
      gate where the same parameter appears.

    Arguments:
        block: parametrized block into which we want to embed parameters.
        to_gate_params: A boolean flag whether to generate gate-level parameters or
            expression-level parameters.

    Returns:
        A tuple with variational parameter dict and the embedding function.
    """

    unique_expressions = unique(expressions(block))
    unique_symbols = [p for p in unique(parameters(block)) if not isinstance(p, sympy.Array)]
    unique_const_matrices = [e for e in unique_expressions if isinstance(e, sympy.Array)]
    unique_expressions = [e for e in unique_expressions if not isinstance(e, sympy.Array)]

    # NOTE
    # there are 3 kinds of parameters in qadence
    # - non-trainable which are considered as inputs for classical data
    # - trainable which are the variational parameters to be optimized
    # - fixed: which are non-trainable parameters with fixed value (e.g. pi/2)
    #
    # both non-trainable and trainable parameters can have the same element applied
    # to different operations in the quantum circuit, e.g. assigning the same parameter
    # to multiple gates.
    non_numeric_symbols = [p for p in unique_symbols if not p.is_number]
    trainable_symbols = [p for p in non_numeric_symbols if p.trainable]
    constant_expressions = [expr for expr in unique_expressions if expr.is_number]
    # we dont need to care about constant symbols if they are contained in an symbolic expression
    # we only care about gate params which are ONLY a constant

    embeddings: dict[sympy.Expr, sympytorch.SymPyModule] = {
        expr: torchify(expr) for expr in unique_expressions if not expr.is_number
    }

    uuid_to_expr = uuid_to_expression(block)

    def embedding_fn(params: StrTensorDict, inputs: StrTensorDict) -> StrTensorDict:
        embedded_params: dict[sympy.Expr, Tensor] = {}
        for expr, fn in embeddings.items():
            angle: Tensor
            values = {}
            for symbol in expr.free_symbols:
                if symbol.name in inputs:
                    value = inputs[symbol.name]
                elif symbol.name in params:
                    value = params[symbol.name]
                else:
                    msg_trainable = "Trainable" if symbol.trainable else "Non-trainable"
                    raise KeyError(
                        f"{msg_trainable} parameter '{symbol.name}' not found in the "
                        f"inputs list: {list(inputs.keys())} nor the "
                        f"params list: {list(params.keys())}."
                    )
                values[symbol.name] = value
            angle = fn(**values)
            # do not reshape parameters which are multi-dimensional
            # tensors, such as for example generator matrices
            if not len(angle.squeeze().shape) > 1:
                angle = angle.reshape(-1)
            embedded_params[expr] = angle

        for e in constant_expressions + unique_const_matrices:
            embedded_params[e] = params[stringify(e)]

        if to_gate_params:
            gate_lvl_params: StrTensorDict = {}
            for uuid, e in uuid_to_expr.items():
                gate_lvl_params[uuid] = embedded_params[e]
            return gate_lvl_params
        else:
            return {stringify(k): v for k, v in embedded_params.items()}

    params: StrTensorDict
    params = {p.name: torch.tensor([p.value], requires_grad=True) for p in trainable_symbols}
    params.update(
        {
            stringify(expr): torch.tensor([evaluate(expr)], requires_grad=False)
            for expr in constant_expressions
        }
    )
    params.update(
        {
            stringify(expr): torch.tensor(
                np.array(expr.tolist(), dtype=np.cdouble), requires_grad=False
            )
            for expr in unique_const_matrices
        }
    )
    return params, embedding_fn



from __future__ import annotations

from typing import Callable

from rich.progress import BarColumn, Progress, TaskProgressColumn, TextColumn, TimeRemainingColumn
from torch import Tensor
from torch.nn import Module
from torch.optim import Optimizer
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter

from qadence.logger import get_logger
from qadence.ml_tools.config import TrainConfig
from qadence.ml_tools.data import DictDataLoader
from qadence.ml_tools.optimize_step import optimize_step
from qadence.ml_tools.printing import print_metrics, write_tensorboard
from qadence.ml_tools.saveload import load_checkpoint, write_checkpoint

logger = get_logger(__name__)


def train(
    model: Module,
    dataloader: DictDataLoader | DataLoader | list[Tensor] | tuple[Tensor, Tensor] | None,
    optimizer: Optimizer,
    config: TrainConfig,
    loss_fn: Callable,
    device: str = "cpu",
    optimize_step: Callable = optimize_step,
    write_tensorboard: Callable = write_tensorboard,
) -> tuple[Module, Optimizer]:
    """Runs the training loop with gradient-based optimizer

    Assumes that `loss_fn` returns a tuple of (loss,
    metrics: dict), where `metrics` is a dict of scalars. Loss and metrics are
    written to tensorboard. Checkpoints are written every
    `config.checkpoint_every` steps (and after the last training step).  If a
    checkpoint is found at `config.folder` we resume training from there.  The
    tensorboard logs can be viewed via `tensorboard --logdir /path/to/folder`.

    Args:
        model: The model to train.
        dataloader: dataloader of different types. If None, no data is required by
            the model
        optimizer: The optimizer to use.
        config: `TrainConfig` with additional training options.
        loss_fn: Loss function returning (loss: float, metrics: dict[str, float])
        device: String defining device to train on, pass 'cuda' for GPU.
        optimize_step: Customizable optimization callback which is called at every iteration.=
            The function must have the signature `optimize_step(model,
            optimizer, loss_fn, xs, device="cpu")` (see the example below).
            Apart from the default we already supply three other optimization
            functions `optimize_step_evo`, `optimize_step_grad_norm`, and
            `optimize_step_inv_dirichlet`. Learn more about how to use this in
            the [Advancded features](../../tutorials/advanced) tutorial of the
            documentation.
        write_tensorboard: Customizable tensorboard logging callback which is
            called every `config.write_every` iterations. The function must have
            the signature `write_tensorboard(writer, loss, metrics, iteration)`
            (see the example below).

    Example:
    ```python exec="on" source="material-block"
    from pathlib import Path
    import torch
    from itertools import count
    from qadence.constructors import hamiltonian_factory, hea, feature_map
    from qadence import chain, Parameter, QuantumCircuit, Z
    from qadence.models import QNN
    from qadence.ml_tools import train_with_grad, TrainConfig

    n_qubits = 2
    fm = feature_map(n_qubits)
    ansatz = hea(n_qubits=n_qubits, depth=3)
    observable = hamiltonian_factory(n_qubits, detuning = Z)
    circuit = QuantumCircuit(n_qubits, fm, ansatz)

    model = QNN(circuit, observable, backend="pyqtorch", diff_mode="ad")
    batch_size = 1
    input_values = {"phi": torch.rand(batch_size, requires_grad=True)}
    pred = model(input_values)

    ## lets prepare the train routine

    cnt = count()
    criterion = torch.nn.MSELoss()
    optimizer = torch.optim.Adam(model.parameters(), lr=0.1)

    def loss_fn(model: torch.nn.Module, data: torch.Tensor) -> tuple[torch.Tensor, dict]:
        next(cnt)
        x, y = data[0], data[1]
        out = model(x)
        loss = criterion(out, y)
        return loss, {}
    tmp_path = Path("/tmp")
    n_epochs = 5
    config = TrainConfig(
        folder=tmp_path,
        max_iter=n_epochs,
        checkpoint_every=100,
        write_every=100,
        batch_size=batch_size,
    )
    batch_size = 25
    x = torch.linspace(0, 1, batch_size).reshape(-1, 1)
    y = torch.sin(x)
    train_with_grad(model, (x, y), optimizer, config, loss_fn=loss_fn)
    ```
    """

    assert loss_fn is not None, "Provide a valid loss function"

    # Move model to device before optimizer is loaded
    model = model.to(device)

    # load available checkpoint
    init_iter = 0
    if config.folder:
        model, optimizer, init_iter = load_checkpoint(config.folder, model, optimizer)
        logger.debug(f"Loaded model and optimizer from {config.folder}")
    # initialize tensorboard
    writer = SummaryWriter(config.folder, purge_step=init_iter)

    ## Training
    progress = Progress(
        TextColumn("[progress.description]{task.description}"),
        BarColumn(),
        TaskProgressColumn(),
        TimeRemainingColumn(elapsed_when_finished=True),
    )
    if isinstance(dataloader, (list, tuple)):
        from qadence.ml_tools.data import to_dataloader

        assert len(dataloader) == 2, "Please provide exactly two torch tensors."
        x, y = dataloader
        dataloader = to_dataloader(x=x, y=y, batch_size=config.batch_size)
    with progress:
        dl_iter = iter(dataloader) if isinstance(dataloader, DictDataLoader) else None

        # outer epoch loop
        for iteration in progress.track(range(init_iter, init_iter + config.max_iter)):
            try:
                # in case there is not data needed by the model
                # this is the case, for example, of quantum models
                # which do not have classical input data (e.g. chemistry)
                if dataloader is None:
                    loss, metrics = optimize_step(
                        model, optimizer, loss_fn, dataloader, device=device
                    )

                # single epoch with DictDataloader using a single iteration method
                # DictDataloader returns a single sample of the data
                # with a given batch size decided when the dataloader is defined
                elif isinstance(dataloader, DictDataLoader):
                    # resample all the time from the dataloader
                    # by creating a fresh iterator if the dataloader
                    # does not support automatically iterating datasets
                    if not dataloader.has_automatic_iter:
                        dl_iter = iter(dataloader)
                    data = next(dl_iter)  # type: ignore[arg-type]
                    loss, metrics = optimize_step(model, optimizer, loss_fn, data, device=device)

                elif isinstance(dataloader, DataLoader):
                    # single-epoch with standard DataLoader
                    # otherwise a standard PyTorch DataLoader behavior
                    # is assumed with optional mini-batches
                    running_loss = 0.0
                    for i, data in enumerate(dataloader):
                        # TODO: make sure to average metrics as well
                        loss, metrics = optimize_step(
                            model, optimizer, loss_fn, data, device=device
                        )
                        running_loss += loss.item()
                    loss = running_loss / (i + 1)

                else:
                    raise NotImplementedError("Unsupported dataloader type!")

                if iteration % config.print_every == 0:
                    print_metrics(loss, metrics, iteration)

                if iteration % config.write_every == 0:
                    write_tensorboard(writer, loss, metrics, iteration)

                if config.folder:
                    if iteration % config.checkpoint_every == 0:
                        write_checkpoint(config.folder, model, optimizer, iteration)

            except KeyboardInterrupt:
                print("Terminating training gracefully after the current iteration.")
                break

    # Final writing and checkpointing
    if config.folder:
        write_checkpoint(config.folder, model, optimizer, iteration)
    write_tensorboard(writer, loss, metrics, iteration)
    writer.close()

    return model, optimizer



from __future__ import annotations

from .errors import NotPauliBlockError, NotSupportedError, QadenceException

# Modules to be automatically added to the qadence namespace
__all__ = []  # type: ignore



from __future__ import annotations

import importlib
from enum import Enum
from typing import Iterable, Tuple, Union

import numpy as np
import sympy
import torch

TNumber = Union[int, float, complex]
"""Union of python number types."""

TDrawColor = Tuple[float, float, float, float]

TParameter = Union[TNumber, torch.Tensor, sympy.Basic, str]
"""Union of numbers, tensors, and parameter types."""

TArray = Union[Iterable, torch.Tensor, np.ndarray]
"""Union of common array types."""

TGenerator = Union[torch.Tensor, sympy.Array, sympy.Basic]
"""Union of torch tensors and numpy arrays."""

PI = torch.pi

# Modules to be automatically added to the qadence namespace
__all__ = [
    "Endianness",
    "Strategy",
    "ResultType",
    "ParameterType",
    "BackendName",
    "StateGeneratorType",
    "LTSOrder",
    "TensorType",
    "DiffMode",
    "BackendName",
    "Interaction",
    "OverlapMethod",
    "AlgoHEvo",
    "SerializationFormat",
]  # type: ignore


class StrEnum(str, Enum):
    def __str__(self) -> str:
        """Used when dumping enum fields in a schema."""
        ret: str = self.value
        return ret

    @classmethod
    def list(cls) -> list[str]:
        return list(map(lambda c: c.value, cls))  # type: ignore


class Strategy(StrEnum):
    """Computing paradigm."""

    DIGITAL = "Digital"
    """Use the digital paradigm."""
    ANALOG = "Analog"
    """Use the analog paradigm."""
    SDAQC = "sDAQC"
    """Use the step-wise digital-analog QC paradigm."""
    BDAQC = "bDAQC"
    """Use the banged digital-analog QC paradigm."""


class Endianness(StrEnum):
    """The endianness convention to use."""

    BIG = "Big"
    """Use Big endianness."""
    LITTLE = "Little"
    """Use little endianness."""


class ResultType(StrEnum):
    """Available data types for generating certain results."""

    STRING = "String"
    """String Type."""
    TORCH = "Torch"
    """Torch Tensor Type."""
    NUMPY = "Numpy"
    """Numpy Array Type."""


class ParameterType(StrEnum):
    """Parameter types available in qadence."""

    FEATURE = "Feature"
    """FeatureParameters act as input and are not trainable."""
    VARIATIONAL = "Variational"
    """VariationalParameters are trainable."""
    FIXED = "Fixed"
    """Fixed/ constant parameters are neither trainable nor act as input."""


class TensorType(StrEnum):
    """Tensor Types for converting blocks to tensors."""

    SPARSEDIAGONAL = "SparseDiagonal"
    """Convert a diagonal observable block to a sparse diagonal if possible."""
    DENSE = "Dense"
    """Convert a block to a dense tensor."""
    SPARSE = "Sparse"
    """Convert a observable block to a sparse tensor."""


class LTSOrder(StrEnum):
    """
    Lie-Trotter-Suzuki approximation order.
    """

    BASIC = "BASIC"
    """Basic."""
    ST2 = "ST2"
    """ST2."""
    ST4 = "ST4"
    """ST4."""


class _DiffMode(StrEnum):
    """Differentiation modes to choose from."""

    GPSR = "gpsr"
    """Basic generalized parameter shift rule."""
    AD = "ad"
    """Automatic Differentiation."""


class QubitSupportType(StrEnum):
    """Qubit support types."""

    GLOBAL = "global"
    """Use global qubit support."""


class Interaction(StrEnum):
    """Interaction types used in
    - [`add_interaction`][qadence.transpile.emulate.add_interaction].
    - [`hamiltonian_factory`][qadence.constructors.hamiltonians.hamiltonian_factory].
    """

    ZZ = "ZZ"
    """ZZ-Ising Interaction"""
    NN = "NN"
    """NN-Ising Interaction, N=(I-Z)/2"""
    XY = "XY"
    """XY Interaction"""
    XYZ = "XYZ"
    """XYZ Interaction"""


class _BackendName(StrEnum):
    """The available backends for running circuits."""

    PYQTORCH = "pyqtorch"
    """The Pyqtorch backend."""
    BRAKET = "braket"
    """The Braket backend."""
    PULSER = "pulser"
    """The Pulser backend."""


# If proprietary qadence_extensions is available, import the
# right function since more backends are supported.
try:
    module = importlib.import_module("qadence_extensions.types")
    BackendName = getattr(module, "BackendName")
    DiffMode = getattr(module, "DiffMode")
except ModuleNotFoundError:
    BackendName = _BackendName
    DiffMode = _DiffMode


class StateGeneratorType(StrEnum):
    """Methods to generate random states."""

    RANDOM_ROTATIONS = "RandomRotations"
    """Random Rotations."""
    HAAR_MEASURE_FAST = "HaarMeasureFast"
    """HaarMeasure."""
    HAAR_MEASURE_SLOW = "HaarMeasureSlow"
    """HaarMeasure non-optimized version."""


class SerializationFormat(StrEnum):
    """Available serialization formats for circuits."""

    PT = "PT"
    """The PT format used by Torch."""
    JSON = "JSON"
    """The Json format."""


class OverlapMethod(StrEnum):
    """Overlap Methods to choose from."""

    EXACT = "exact"
    """Exact."""
    JENSEN_SHANNON = "jensen_shannon"
    """Jensen-shannon."""
    COMPUTE_UNCOMPUTE = "compute_uncompute"
    """Compute-uncompute."""
    SWAP_TEST = "swap_test"
    """Swap-test."""
    HADAMARD_TEST = "hadamard_test"
    """Hadamard-test."""


class FigFormat(StrEnum):
    """Available output formats for exporting visualized circuits to a file."""

    PNG = "PNG"
    """PNG format."""
    PDF = "PDF"
    """PDF format."""
    SVG = "SVG"
    """SVG format."""


class AlgoHEvo(StrEnum):
    """Hamiltonian Evolution algorithms that can be used by the backend."""

    RK4 = "RK4"
    """4th order Runge-Kutta approximation."""
    EIG = "EIG"
    """Using Hamiltonian diagonalization."""
    EXP = "EXP"
    """Using torch.matrix_exp on the generator matrix."""


class LatticeTopology(StrEnum):
    """Lattice topologies to choose from for the register."""

    LINE = "line"
    """Line-format lattice."""
    SQUARE = "square"
    """Square lattice."""
    CIRCLE = "circle"
    """Circular lattice."""
    ALL_TO_ALL = "all_to_all"
    """All to all- connected lattice."""
    RECTANGULAR_LATTICE = "rectangular_lattice"
    """Rectangular-shaped lattice."""
    TRIANGULAR_LATTICE = "triangular_lattice"
    """Triangular-shaped shape."""
    HONEYCOMB_LATTICE = "honeycomb_lattice"
    """Honeycomb-shaped lattice."""
    ARBITRARY = "arbitrary"
    """Arbitrarily-shaped lattice."""


class GenDAQC(StrEnum):
    """The type of interaction for the DAQC transform."""

    ZZ = "ZZ"
    """ZZ"""
    NN = "NN"
    """NN"""


class OpName(StrEnum):
    """A list of all available of digital-analog operations."""

    # Digital operations
    X = "X"
    """The X gate."""
    Y = "Y"
    """The Y gate."""
    Z = "Z"
    """The Z gate."""
    N = "N"
    """The N = (1/2)(I-Z) operator"""
    H = "H"
    """The Hadamard gate."""
    I = "I"  # noqa
    """The Identity gate."""
    ZERO = "Zero"
    """The zero gate."""
    RX = "RX"
    """The RX gate."""
    RY = "RY"
    """The RY gate."""
    RZ = "RZ"
    """The RZ gate."""
    U = "U"
    """The U gate."""
    CNOT = "CNOT"
    """The CNOT gate."""
    CZ = "CZ"
    """The CZ gate."""
    MCZ = "MCZ"
    """The Multicontrol CZ gate."""
    HAMEVO = "HamEvo"
    """The Hamiltonian Evolution operation."""
    CRX = "CRX"
    """The Control RX gate."""
    MCRX = "MCRX"
    """The Multicontrol RX gate."""
    CRY = "CRY"
    """The Controlled RY gate."""
    MCRY = "MCRY"
    """The Multicontrol RY gate."""
    CRZ = "CRZ"
    """The Control RZ gate."""
    MCRZ = "MCRZ"
    """The Multicontrol RZ gate."""
    CSWAP = "CSWAP"
    """The Control SWAP gate."""
    T = "T"
    """The T gate."""
    # FIXME: Tdagger is not currently supported by any backend
    TDAGGER = "TDagger"
    """The T dagger gate."""
    S = "S"
    """The S gate."""
    SDAGGER = "SDagger"
    """The S dagger gate."""
    SWAP = "SWAP"
    """The SWAP gate."""
    PHASE = "PHASE"
    """The PHASE gate."""
    CPHASE = "CPHASE"
    """The controlled PHASE gate."""
    MCPHASE = "MCPHASE"
    """The Multicontrol PHASE gate."""
    TOFFOLI = "Toffoli"
    """The Toffoli gate."""
    # Analog operations
    ANALOGENTANG = "AnalogEntanglement"
    """The analog entanglement operation."""
    ANALOGRX = "AnalogRX"
    """The analog RX operation."""
    ANALOGRY = "AnalogRY"
    """The analog RY operation."""
    ANALOGRZ = "AnalogRZ"
    """The analog RZ operation."""
    ANALOGSWAP = "AnalogSWAP"
    """The analog SWAP operation."""
    ENTANG = "entangle"
    """The entanglement operation."""
    WAIT = "wait"
    """The wait operation."""



from __future__ import annotations

from qadence.backend import Backend, BackendConfiguration
from qadence.backends.pytorch_wrapper import DifferentiableBackend, DiffMode
from qadence.extensions import available_backends, set_backend_config
from qadence.types import BackendName

__all__ = ["backend_factory", "config_factory"]


def backend_factory(
    backend: BackendName | str,
    diff_mode: DiffMode | str | None = None,
    configuration: BackendConfiguration | dict | None = None,
) -> Backend | DifferentiableBackend:
    backend_inst: Backend | DifferentiableBackend
    backend_name = BackendName(backend)
    backends = available_backends()

    try:
        BackendCls = backends[backend_name]
    except (KeyError, ValueError):
        raise NotImplementedError(f"The requested backend '{backend_name}' is not implemented.")

    default_config = BackendCls.default_configuration()
    if configuration is None:
        configuration = default_config
    elif isinstance(configuration, dict):
        configuration = config_factory(backend_name, configuration)
    else:
        # NOTE: types have to match exactly, hence we use `type`
        if not isinstance(configuration, type(BackendCls.default_configuration())):
            raise ValueError(
                f"Given config class '{type(configuration)}' does not match the backend",
                f" class: '{BackendCls}'. Expected: '{type(BackendCls.default_configuration())}.'",
            )

    # Create the backend
    backend_inst = BackendCls(
        config=configuration
        if configuration is not None
        else BackendCls.default_configuration()  # type: ignore[attr-defined]
    )

    # Set backend configurations which depend on the differentiation mode
    set_backend_config(backend_inst, diff_mode)

    if diff_mode is not None:
        backend_inst = DifferentiableBackend(backend_inst, DiffMode(diff_mode))
    return backend_inst


def config_factory(name: BackendName | str, config: dict) -> BackendConfiguration:
    backends = available_backends()

    try:
        BackendCls = backends[BackendName(name)]
    except KeyError:
        raise NotImplementedError(f"The requested backend '{name}' is not implemented!")

    BackendConfigCls = type(BackendCls.default_configuration())
    return BackendConfigCls(**config)  # type: ignore[no-any-return]



from __future__ import annotations

from typing import List

from qadence.blocks.utils import chain
from qadence.circuit import QuantumCircuit
from qadence.operations import I


def reverse_program(circuit: QuantumCircuit, inplace: bool = True) -> QuantumCircuit:
    raise (NotImplementedError("Must also implement at Block level"))


def parametric_shift_rule(circuit: QuantumCircuit) -> List[QuantumCircuit]:
    raise NotImplementedError


def fill_identities(circ: QuantumCircuit) -> QuantumCircuit:
    empty_wires = set(range(circ.n_qubits)) - set(circ.block.qubit_support)
    if len(empty_wires) > 0:
        ids = chain(I(i) for i in empty_wires)
        return QuantumCircuit(circ.n_qubits, chain(circ.block, ids))
    return circ



from __future__ import annotations

from copy import deepcopy
from functools import singledispatch
from itertools import product
from math import dist as euclidean_distance
from typing import Any, Callable, Union, overload

import torch
from sympy import cos, sin

from qadence.blocks.abstract import AbstractBlock
from qadence.blocks.analog import (
    AnalogBlock,
    AnalogChain,
    AnalogKron,
    ConstantAnalogRotation,
    Interaction,
    WaitBlock,
)
from qadence.blocks.composite import CompositeBlock
from qadence.blocks.primitive import PrimitiveBlock, ScaleBlock
from qadence.blocks.utils import _construct
from qadence.circuit import QuantumCircuit
from qadence.operations import HamEvo, I, N, X, Y, add, chain, kron, wait
from qadence.qubit_support import QubitSupport
from qadence.register import Register
from qadence.transpile.transpile import blockfn_to_circfn

C6_DICT = {
    50: 96120.72,
    51: 122241.6,
    52: 154693.02,
    53: 194740.36,
    54: 243973.91,
    55: 304495.01,
    56: 378305.98,
    57: 468027.05,
    58: 576714.85,
    59: 707911.38,
    60: 865723.02,
    61: 1054903.11,
    62: 1281042.11,
    63: 1550531.15,
    64: 1870621.31,
    65: 2249728.57,
    66: 2697498.69,
    67: 3224987.51,
    68: 3844734.37,
    69: 4571053.32,
    70: 5420158.53,
    71: 6410399.4,
    72: 7562637.31,
    73: 8900342.14,
    74: 10449989.62,
    75: 12241414.53,
    76: 14308028.03,
    77: 16687329.94,
    78: 19421333.62,
    79: 22557029.94,
    80: 26146720.74,
    81: 30248886.65,
    82: 34928448.69,
    83: 40257623.67,
    84: 46316557.88,
    85: 53194043.52,
    86: 60988354.64,
    87: 69808179.15,
    88: 79773468.88,
    89: 91016513.07,
    90: 103677784.57,
    91: 117933293.96,
    92: 133943541.9,
    93: 151907135.94,
    94: 172036137.34,
    95: 194562889.89,
    96: 219741590.56,
    97: 247850178.91,
    98: 279192193.77,
    99: 314098829.39,
    100: 352931119.11,
}


def _qubitposition(register: Register, i: int) -> tuple[int, int]:
    (x, y) = list(register.coords.values())[i]
    return (x, y)


def ising_interaction(
    register: Register, pairs: list[tuple[int, int]], rydberg_level: int = 60
) -> AbstractBlock:
    c6 = C6_DICT[rydberg_level]

    def term(i: int, j: int) -> AbstractBlock:
        qi, qj = _qubitposition(register, i), _qubitposition(register, j)
        rij = euclidean_distance(qi, qj)
        return (c6 / rij**6) * kron(N(i), N(j))

    return add(term(i, j) for (i, j) in pairs)


def xy_interaction(
    register: Register, pairs: list[tuple[int, int]], c3: float = 3700.0
) -> AbstractBlock:
    def term(i: int, j: int) -> AbstractBlock:
        qi, qj = _qubitposition(register, i), _qubitposition(register, j)
        rij = euclidean_distance(qi, qj)
        return (c3 / rij**3) * (kron(X(i), X(j)) + kron(Y(i), Y(j)))

    return add(term(i, j) for (i, j) in pairs)


INTERACTIONS = {Interaction.NN: ising_interaction, Interaction.XY: xy_interaction}


@overload
def add_interaction(circuit: QuantumCircuit, **kwargs: Any) -> QuantumCircuit:
    ...


@overload
def add_interaction(block: AbstractBlock, **kwargs: Any) -> AbstractBlock:
    ...


@overload
def add_interaction(register: Register, block: AbstractBlock, **kwargs: Any) -> AbstractBlock:
    ...


@singledispatch
def add_interaction(
    x: Register | QuantumCircuit | AbstractBlock,
    *args: Any,
    interaction: Interaction | Callable = Interaction.NN,
    spacing: float = 1.0,
) -> QuantumCircuit | AbstractBlock:
    """Turns blocks or circuits into (a chain of) `HamEvo` blocks including a
    chosen interaction term.

    This is a `@singledipatch`ed function which can be called in three ways:

    * With a `QuantumCircuit` which contains all necessary information: `add_interaction(circuit)`
    * With a `Register` and an `AbstractBlock`: `add_interaction(reg, block)`
    * With an `AbstractBlock` only: `add_interaction(block)`

    See the section about [analog blocks](/digital_analog_qc/analog-basics.md) for
    detailed information about how which types of blocks are translated.

    Arguments:
        x: Circuit or block to be emulated. See the examples on which argument
            combinations are accepted.
        interaction: Type of interaction that is added. Can also be a function that accepts a
            register and a list of edges that define which qubits interact (see the examples).
        spacing: All qubit coordinates are multiplied by `spacing`.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence import QuantumCircuit, AnalogRX, add_interaction

    c = QuantumCircuit(2, AnalogRX(2.0))
    e = add_interaction(c)
    print(str(e.block.generator)) # markdown-exec: hide
    ```
    You can also use `add_interaction` directly on a block, but you have to provide either
    the `Register` or define a non-global qubit support.
    ```python exec="on" source="material-block" result="json"
    from qadence import AnalogRX, Register, add_interaction

    b = AnalogRX(2.0)
    r = Register(1)
    e = add_interaction(r, b)
    print(e.generator) # markdown-exec: hide

    # or provide only the block with local qubit support
    # in this case the register is created via `Register(b.n_qubits)`
    e = add_interaction(AnalogRX(2.0, qubit_support=(0,)))
    print(e.generator)
    ```
    You can specify a custom `interaction` function which has to accept a `Register` and a list
    of `edges: list[tuple[int, int]]`:
    ```python exec="on" source="material-block" result="json"
    from qadence import AnalogRX, Register, add_interaction
    from qadence.transpile.emulate import ising_interaction

    def int_fn(r: Register, pairs: list[tuple[int, int]]) -> AbstractBlock:
        # do either something completely custom
        # ...
        # or e.g. change the default kwargs to `ising_interaction`
        return ising_interaction(r, pairs, rydberg_level=70)

    b = AnalogRX(2.0)
    r = Register(1)
    e = add_interaction(r, b, interaction=int_fn)
    ```
    """
    raise ValueError(f"`add_interaction` is not implemented for {type(x)}")


@add_interaction.register  # type: ignore[attr-defined]
def _(circuit: QuantumCircuit, **kwargs: Any) -> QuantumCircuit:
    reg = circuit.register
    return blockfn_to_circfn(lambda b: add_interaction(reg, b, **kwargs))(circuit)


@add_interaction.register  # type: ignore[attr-defined]
def _(block: AbstractBlock, **kwargs: Any) -> AbstractBlock:
    return add_interaction(Register(block.n_qubits), block, **kwargs)


@add_interaction.register  # type: ignore[attr-defined]
def _(
    register: Register,
    block: AbstractBlock,
    interaction: Union[Interaction, Callable] = Interaction.NN,
    spacing: float = 1.0,
) -> AbstractBlock:
    try:
        fn = interaction if callable(interaction) else INTERACTIONS[Interaction(interaction)]
    except KeyError:
        raise KeyError(
            "Function `add_interaction` only supports NN and XY, or a custom callable function."
        )
    reg = register._scale_positions(spacing)
    return _add_interaction(block, reg, fn)  # type: ignore[arg-type]


@singledispatch
def _add_interaction(b: AbstractBlock, r: Register, interaction: Callable) -> AbstractBlock:
    raise NotImplementedError(f"Cannot emulate {type(b)}")


@_add_interaction.register
def _(b: CompositeBlock, r: Register, i: Callable) -> AbstractBlock:
    return _construct(type(b), tuple(map(lambda b: _add_interaction(b, r, i), b.blocks)))


@_add_interaction.register
def _(block: ScaleBlock, register: Register, interaction: Callable) -> AbstractBlock:
    if isinstance(block.block, AnalogBlock):
        raise NotImplementedError("Scaling emulated analog blocks is not implemented.")
    return block


@_add_interaction.register
def _(block: PrimitiveBlock, register: Register, interaction: Callable) -> AbstractBlock:
    return block


@_add_interaction.register
def _(block: WaitBlock, register: Register, interaction: Callable) -> AbstractBlock:
    duration = block.parameters.duration

    support = tuple(range(register.n_qubits))
    assert support == block.qubit_support if not block.qubit_support.is_global else True
    pairs = list(filter(lambda x: x[0] < x[1], product(support, support)))

    return HamEvo(interaction(register, pairs), duration / 1000) if len(pairs) else I(0)


def rot_generator(block: ConstantAnalogRotation) -> AbstractBlock:
    omega = block.parameters.omega
    delta = block.parameters.delta
    phase = block.parameters.phase
    support = block.qubit_support

    x_terms = (omega / 2) * add(cos(phase) * X(i) - sin(phase) * Y(i) for i in support)
    z_terms = delta * add(N(i) for i in support)
    return x_terms - z_terms  # type: ignore[no-any-return]


@_add_interaction.register
def _(block: ConstantAnalogRotation, register: Register, interaction: Callable) -> AbstractBlock:
    # convert "global" to indexed qubit suppport so that we can re-use `kron` dispatched function
    b = deepcopy(block)
    b.qubit_support = QubitSupport(*range(register.n_qubits))
    return _add_interaction(kron(b), register, interaction)


@_add_interaction.register
def _(block: AnalogKron, register: Register, interaction: Callable) -> AbstractBlock:
    from qadence import block_to_tensor

    w_block = wait(duration=block.duration, qubit_support=block.qubit_support)
    i_terms = add_interaction(register, w_block, interaction=interaction)

    generator = add(rot_generator(b) for b in block.blocks if isinstance(b, ConstantAnalogRotation))
    generator = generator if i_terms == I(0) else generator + i_terms.generator  # type: ignore[attr-defined]  # noqa: E501

    norm = torch.norm(block_to_tensor(generator)).item()
    return HamEvo(generator / norm, norm * block.duration / 1000)


@_add_interaction.register
def _(block: AnalogChain, register: Register, interaction: Callable) -> AbstractBlock:
    return chain(add_interaction(register, b, interaction=interaction) for b in block.blocks)



from __future__ import annotations

from typing import Iterable, Type

import numpy as np
import sympy

from qadence.blocks import KronBlock, kron
from qadence.operations import RY
from qadence.parameters import FeatureParameter, Parameter


def generator_prefactor(spectrum: str, qubit_index: int) -> float | int:
    """
    Converts a spectrum string (e.g., tower or exponential) to the correct generator prefactor.
    """
    spectrum = spectrum.lower()
    conversion_dict: dict[str, float | int] = {
        "simple": 1,
        "tower": qubit_index + 1,
        "exponential": 2 * np.pi / (2 ** (qubit_index + 1)),
    }
    return conversion_dict[spectrum]


def basis_func(basis: str, x: Parameter) -> Parameter | sympy.Expr:
    basis = basis.lower()
    conversion_dict: dict[str, Parameter | sympy.Expr] = {
        "fourier": x,
        "chebyshev": 2 * sympy.acos(x),
    }
    return conversion_dict[basis]


def build_idx_fms(
    basis: str,
    fm_pauli: Type[RY],
    fm_strategy: str,
    n_features: int,
    n_qubits: int,
    spectrum: str,
) -> list[KronBlock]:
    """Builds the index feature maps based on the given parameters.

    Args:
        basis (str): Type of basis chosen for the feature map.
        fm_pauli (PrimitiveBlock type): The chosen Pauli rotation type.
        fm_strategy (str): The feature map strategy to be used. Possible values are
            'parallel' or 'serial'.
        n_features (int): The number of features.
        n_qubits (int): The number of qubits.
        spectrum (str): The chosen spectrum.

    Returns:
        List[KronBlock]: The list of index feature maps.
    """
    idx_fms = []
    for i in range(n_features):
        target_qubits = get_fm_qubits(fm_strategy, i, n_qubits, n_features)
        param = FeatureParameter(f"x{i}")
        block = kron(
            *[
                fm_pauli(qubit, generator_prefactor(spectrum, j) * basis_func(basis, param))
                for j, qubit in enumerate(target_qubits)
            ]
        )
        idx_fm = block
        idx_fms.append(idx_fm)
    return idx_fms


def get_fm_qubits(fm_strategy: str, i: int, n_qubits: int, n_features: int) -> Iterable:
    """Returns the list of target qubits for the given feature map strategy and feature index

    Args:
        fm_strategy (str): The feature map strategy to be used. Possible values
            are 'parallel' or 'serial'.
        i (int): The feature index.
        n_qubits (int): The number of qubits.
        n_features (int): The number of features.

    Returns:
        List[int]: The list of target qubits.

    Raises:
        ValueError: If the feature map strategy is not implemented.
    """
    if fm_strategy == "parallel":
        n_qubits_per_feature = int(n_qubits / n_features)
        target_qubits = range(i * n_qubits_per_feature, (i + 1) * n_qubits_per_feature)
    elif fm_strategy == "serial":
        target_qubits = range(0, n_qubits)
    else:
        raise ValueError(f"Feature map strategy {fm_strategy} not implemented.")
    return target_qubits



from __future__ import annotations

from typing import Type, Union

import numpy as np
import sympy

from qadence.blocks import AbstractBlock, KronBlock, chain, kron, tag
from qadence.operations import RX, RY, RZ, H
from qadence.parameters import FeatureParameter, Parameter

Rotation = Union[RX, RY, RZ]


def feature_map(
    n_qubits: int,
    support: tuple[int, ...] = None,
    param: str = "phi",
    op: Type[Rotation] = RX,
    fm_type: str = "fourier",
) -> KronBlock:
    """Construct a feature map of a given type.

    Arguments:
        n_qubits: Number of qubits the feature map covers. Results in `support=range(n_qubits)`.
        support: Overrides `n_qubits`. Puts one rotation gate on every qubit in `support`.
        param: Parameter of the feature map.
        op: Rotation operation of the feature map.
        fm_type: Determines the additional expression the final feature parameter (the addtional
            term in front of `param`). `"fourier": param` (nothing is done to `param`)
            `"chebyshev": 2*acos(param)`, `"tower": (i+1)*2*acos(param)` (where `i` is the qubit
            index).

    Example:
    ```python exec="on" source="material-block" result="json"
    from qadence import feature_map

    fm = feature_map(3, fm_type="fourier")
    print(f"{fm = }")

    fm = feature_map(3, fm_type="chebyshev")
    print(f"{fm = }")

    fm = feature_map(3, fm_type="tower")
    print(f"{fm = }")
    ```
    """
    fparam = FeatureParameter(param)
    if support is None:
        support = tuple(range(n_qubits))

    assert len(support) <= n_qubits, "Wrong qubit support supplied"

    if fm_type == "fourier":
        fm = kron(*[op(qubit, fparam) for qubit in support])
    elif fm_type == "chebyshev":
        fm = kron(*[op(qubit, 2 * sympy.acos(fparam)) for qubit in support])
    elif fm_type == "tower":
        fm = kron(*[op(qubit, (i + 1) * 2 * sympy.acos(fparam)) for i, qubit in enumerate(support)])
    else:
        raise NotImplementedError(f"Feature map {fm_type} not implemented")
    fm.tag = "FM"
    return fm


def fourier_feature_map(
    n_qubits: int, support: tuple[int, ...] = None, param: str = "phi", op: Type[Rotation] = RX
) -> AbstractBlock:
    """Construct a Fourier feature map

    Args:
        n_qubits: number of qubits across which the FM is created
        param: The base name for the feature `Parameter`
    """
    fm = feature_map(n_qubits, support=support, param=param, op=op, fm_type="fourier")
    return tag(fm, tag="FourierFM")


def chebyshev_feature_map(
    n_qubits: int, support: tuple[int, ...] = None, param: str = "phi", op: Type[Rotation] = RX
) -> AbstractBlock:
    """Construct a Chebyshev feature map

    Args:
        n_qubits: number of qubits across which the FM is created
        support (Iterable[int]): The qubit support
        param: The base name for the feature `Parameter`
    """
    fm = feature_map(n_qubits, support=support, param=param, op=op, fm_type="chebyshev")
    return tag(fm, tag="ChebyshevFM")


def tower_feature_map(
    n_qubits: int, support: tuple[int, ...] = None, param: str = "phi", op: Type[Rotation] = RX
) -> AbstractBlock:
    """Construct a Chebyshev tower feature map

    Args:
        n_qubits: number of qubits across which the FM is created
        param: The base name for the feature `Parameter`
    """
    fm = feature_map(n_qubits, support=support, param=param, op=op, fm_type="tower")
    return tag(fm, tag="TowerFM")


def exp_fourier_feature_map(
    n_qubits: int,
    support: tuple[int, ...] = None,
    param: str = "x",
    feature_range: tuple[float, float] = None,
) -> AbstractBlock:
    """
    Exponential fourier feature map.

    Args:
        n_qubits: number of qubits in the feature
        support: qubit support
        param: name of feature `Parameter`
        feature_range: min and max value of the feature, as floats in a Tuple
    """

    if feature_range is None:
        feature_range = (0.0, 2.0**n_qubits)

    if support is None:
        support = tuple(range(n_qubits))

    xmax = max(feature_range)
    xmin = min(feature_range)

    x = Parameter(param, trainable=False)

    # The feature map works on the range of 0 to 2**n
    x_rescaled = 2 * np.pi * (x - xmin) / (xmax - xmin)

    hlayer = kron(H(qubit) for qubit in support)
    rlayer = kron(RZ(support[i], x_rescaled * (2**i)) for i in range(n_qubits))

    return tag(chain(hlayer, rlayer), f"ExpFourierFM({param})")
